# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import HealthCheck, given, settings, strategies
import pytest

from swh.indexer.metadata_dictionary import MAPPINGS


def test_gemspec_base():
    raw_content = b"""
Gem::Specification.new do |s|
s.name        = 'example'
s.version     = '0.1.0'
s.licenses    = ['MIT']
s.summary     = "This is an example!"
s.description = "Much longer explanation of the example!"
s.authors     = ["Ruby Coder"]
s.email       = 'rubycoder@example.com'
s.files       = ["lib/example.rb"]
s.homepage    = 'https://rubygems.org/gems/example'
s.metadata    = { "source_code_uri" => "https://github.com/example/example" }
end"""
    result = MAPPINGS["GemspecMapping"]().translate(raw_content)
    assert set(result.pop("description")) == {
        "This is an example!",
        "Much longer explanation of the example!",
    }
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [{"type": "Person", "name": "Ruby Coder"}],
        "name": "example",
        "license": "https://spdx.org/licenses/MIT",
        "codeRepository": "https://rubygems.org/gems/example",
        "email": "rubycoder@example.com",
        "version": "0.1.0",
    }


@pytest.mark.xfail(reason="https://github.com/w3c/json-ld-api/issues/547")
def test_gemspec_two_author_fields():
    raw_content = b"""
Gem::Specification.new do |s|
s.authors     = ["Ruby Coder1"]
s.author      = "Ruby Coder2"
end"""
    result = MAPPINGS["GemspecMapping"]().translate(raw_content)
    assert result.pop("author") in (
        [
            {"type": "Person", "name": "Ruby Coder1"},
            {"type": "Person", "name": "Ruby Coder2"},
        ],
        [
            {"type": "Person", "name": "Ruby Coder2"},
            {"type": "Person", "name": "Ruby Coder1"},
        ],
    )
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }


def test_gemspec_invalid_author():
    raw_content = b"""
Gem::Specification.new do |s|
s.author      = ["Ruby Coder"]
end"""
    result = MAPPINGS["GemspecMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }
    raw_content = b"""
Gem::Specification.new do |s|
s.author      = "Ruby Coder1",
end"""
    result = MAPPINGS["GemspecMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }
    raw_content = b"""
Gem::Specification.new do |s|
s.authors     = ["Ruby Coder1", ["Ruby Coder2"]]
end"""
    result = MAPPINGS["GemspecMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [{"type": "Person", "name": "Ruby Coder1"}],
    }


def test_gemspec_alternative_header():
    raw_content = b"""
require './lib/version'

Gem::Specification.new { |s|
s.name = 'rb-system-with-aliases'
s.summary = 'execute system commands with aliases'
}
"""
    result = MAPPINGS["GemspecMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "rb-system-with-aliases",
        "description": "execute system commands with aliases",
    }


@settings(suppress_health_check=[HealthCheck.too_slow])
@given(
    strategies.dictionaries(
        # keys
        strategies.one_of(
            strategies.text(),
            *map(strategies.just, MAPPINGS["GemspecMapping"].mapping),  # type: ignore
        ),
        # values
        strategies.recursive(
            strategies.characters(),
            lambda children: strategies.lists(children, min_size=1),
        ),
    )
)
def test_gemspec_adversarial(doc):
    parts = [b"Gem::Specification.new do |s|\n"]
    for k, v in doc.items():
        parts.append("  s.{} = {}\n".format(k, repr(v)).encode())
    parts.append(b"end\n")
    MAPPINGS["GemspecMapping"]().translate(b"".join(parts))



# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from uuid import uuid1

from hypothesis.strategies import composite, one_of, sampled_from, sets, tuples, uuids

from swh.model.hashutil import MultiHash

MIMETYPES = [
    b"application/json",
    b"application/octet-stream",
    b"application/xml",
    b"text/plain",
]

ENCODINGS = [
    b"iso8859-1",
    b"iso8859-15",
    b"latin1",
    b"utf-8",
]


def gen_mimetype():
    """Generate one mimetype strategy."""
    return one_of(sampled_from(MIMETYPES))


def gen_encoding():
    """Generate one encoding strategy."""
    return one_of(sampled_from(ENCODINGS))


def _init_content(uuid):
    """Given a uuid, initialize a content"""
    return {
        "id": MultiHash.from_data(uuid.bytes, {"sha1"}).digest()["sha1"],
        "indexer_configuration_id": 1,
    }


@composite
def gen_content_mimetypes(draw, *, min_size=0, max_size=100):
    """Generate valid and consistent content_mimetypes.

    Context: Test purposes

    Args:
        **draw** (callable): Used by hypothesis to generate data
        **min_size** (int): Minimal number of elements to generate
                            (default: 0)
        **max_size** (int): Maximal number of elements to generate
                            (default: 100)

    Returns:
        List of content_mimetypes as expected by the
        content_mimetype_add api endpoint.

    """
    _ids = draw(
        sets(
            tuples(uuids(), gen_mimetype(), gen_encoding()),
            min_size=min_size,
            max_size=max_size,
        )
    )

    content_mimetypes = []
    for uuid, mimetype, encoding in _ids:
        content_mimetypes.append(
            {
                **_init_content(uuid),
                "mimetype": mimetype,
                "encoding": encoding,
            }
        )
    return content_mimetypes


TOOLS = [
    {
        "tool_name": "swh-metadata-translator",
        "tool_version": "0.0.1",
        "tool_configuration": {"type": "local", "context": "NpmMapping"},
    },
    {
        "tool_name": "swh-metadata-detector",
        "tool_version": "0.0.1",
        "tool_configuration": {
            "type": "local",
            "context": ["NpmMapping", "CodemetaMapping"],
        },
    },
    {
        "tool_name": "swh-metadata-detector2",
        "tool_version": "0.0.1",
        "tool_configuration": {
            "type": "local",
            "context": ["NpmMapping", "CodemetaMapping"],
        },
    },
    {
        "tool_name": "file",
        "tool_version": "5.22",
        "tool_configuration": {"command_line": "file --mime <filepath>"},
    },
    {
        "tool_name": "pygments",
        "tool_version": "2.0.1+dfsg-1.1+deb8u1",
        "tool_configuration": {"type": "library", "debian-package": "python3-pygments"},
    },
    {
        "tool_name": "pygments2",
        "tool_version": "2.0.1+dfsg-1.1+deb8u1",
        "tool_configuration": {
            "type": "library",
            "debian-package": "python3-pygments",
            "max_content_size": 10240,
        },
    },
    {
        "tool_name": "nomos",
        "tool_version": "3.1.0rc2-31-ga2cbb8c",
        "tool_configuration": {"command_line": "nomossa <filepath>"},
    },
]


MIMETYPE_OBJECTS = [
    {
        "id": MultiHash.from_data(uuid1().bytes, {"sha1"}).digest()["sha1"],
        "mimetype": mt,
        "encoding": enc,
        # 'indexer_configuration_id' will be added after TOOLS get registered
    }
    for mt in MIMETYPES
    for enc in ENCODINGS
]

LICENSES = [
    b"3DFX",
    b"BSD",
    b"GPL",
    b"Apache2",
    b"MIT",
]

FOSSOLOGY_LICENSES = [
    {
        "id": MultiHash.from_data(uuid1().bytes, {"sha1"}).digest()["sha1"],
        "licenses": [
            LICENSES[i % len(LICENSES)],
        ],
        # 'indexer_configuration_id' will be added after TOOLS get registered
    }
    for i in range(10)
]


def gen_license():
    return one_of(sampled_from(LICENSES))


@composite
def gen_content_fossology_licenses(draw, *, min_size=0, max_size=100):
    """Generate valid and consistent content_fossology_licenses.

    Context: Test purposes

    Args:
        **draw** (callable): Used by hypothesis to generate data
        **min_size** (int): Minimal number of elements to generate
                            (default: 0)
        **max_size** (int): Maximal number of elements to generate
                            (default: 100)

    Returns:
        List of content_fossology_licenses as expected by the
        content_fossology_license_add api endpoint.

    """
    _ids = draw(
        sets(
            tuples(
                uuids(),
                gen_license(),
            ),
            min_size=min_size,
            max_size=max_size,
        )
    )

    content_licenses = []
    for uuid, license in _ids:
        content_licenses.append(
            {
                **_init_content(uuid),
                "licenses": [license],
            }
        )
    return content_licenses



# Copyright (C) 2015-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import math
import threading
from typing import Any, Dict, List, Tuple, Type

import attr
import pytest

from swh.indexer.storage.exc import DuplicateId, IndexerStorageArgumentException
from swh.indexer.storage.interface import IndexerStorageInterface, PagedResult
from swh.indexer.storage.model import (
    BaseRow,
    ContentLicenseRow,
    ContentMetadataRow,
    ContentMimetypeRow,
    DirectoryIntrinsicMetadataRow,
    OriginExtrinsicMetadataRow,
    OriginIntrinsicMetadataRow,
)
from swh.model.hashutil import hash_to_bytes


def _remove_tool_ids(rows):
    results = []
    for row in rows:
        tool = dict(row.tool)
        del tool["id"]
        results.append(attr.evolve(row, tool=tool))
    return results


def prepare_mimetypes_from_licenses(
    fossology_licenses: List[ContentLicenseRow],
) -> List[ContentMimetypeRow]:
    """Fossology license needs some consistent data in db to run."""
    mimetypes = []
    for c in fossology_licenses:
        mimetypes.append(
            ContentMimetypeRow(
                id=c.id,
                mimetype="text/plain",  # for filtering on textual data to work
                encoding="utf-8",
                indexer_configuration_id=c.indexer_configuration_id,
            )
        )
    return mimetypes


def endpoint_name(etype: str, ename: str) -> str:
    """Compute the storage's endpoint's name

    >>> endpoint_name('content_mimetype', 'add')
    'content_mimetype_add'
    >>> endpoint_name('content_fosso_license', 'delete')
    'content_fosso_license_delete'

    """
    return f"{etype}_{ename}"


def endpoint(storage, etype: str, ename: str):
    return getattr(storage, endpoint_name(etype, ename))


def expected_summary(count: int, etype: str, ename: str = "add") -> Dict[str, int]:
    """Compute the expected summary

    The key is determine according to etype and ename

        >>> expected_summary(10, 'content_mimetype', 'add')
        {'content_mimetype:add': 10}
        >>> expected_summary(9, 'origin_intrinsic_metadata', 'delete')
        {'origin_intrinsic_metadata:del': 9}

    """
    pattern = ename[0:3]
    key = endpoint_name(etype, ename).replace(f"_{ename}", f":{pattern}")
    return {key: count}


def test_check_config(swh_indexer_storage) -> None:
    assert swh_indexer_storage.check_config(check_write=True)
    assert swh_indexer_storage.check_config(check_write=False)


class StorageETypeTester:
    """Base class for testing a series of common behaviour between a bunch of
    endpoint types supported by an IndexerStorage.

    This is supposed to be inherited with the following class attributes:
    - endpoint_type
    - tool_name
    - example_data

    See below for example usage.
    """

    endpoint_type: str
    tool_name: str
    example_data: List[Dict]
    row_class: Type[BaseRow]

    def test_missing(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool_id = data.tools[self.tool_name]["id"]

        # given 2 (hopefully) unknown objects
        query = [
            {
                "id": data.sha1_1,
                "indexer_configuration_id": tool_id,
            },
            {
                "id": data.sha1_2,
                "indexer_configuration_id": tool_id,
            },
        ]

        # we expect these are both returned by the xxx_missing endpoint
        actual_missing = endpoint(storage, etype, "missing")(query)
        assert list(actual_missing) == [
            data.sha1_1,
            data.sha1_2,
        ]

        # now, when we add one of them
        summary = endpoint(storage, etype, "add")(
            [
                self.row_class.from_dict(
                    {
                        "id": data.sha1_2,
                        **self.example_data[0],
                        "indexer_configuration_id": tool_id,
                    }
                )
            ]
        )

        assert summary == expected_summary(1, etype)

        # we expect only the other one returned
        actual_missing = endpoint(storage, etype, "missing")(query)
        assert list(actual_missing) == [data.sha1_1]

    def test_add__update_in_place_duplicate(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        data_v1 = {
            "id": data.sha1_2,
            **self.example_data[0],
            "indexer_configuration_id": tool["id"],
        }

        # given
        summary = endpoint(storage, etype, "add")([self.row_class.from_dict(data_v1)])
        assert summary == expected_summary(1, etype)  # not added

        # when
        actual_data = list(endpoint(storage, etype, "get")([data.sha1_2]))

        expected_data_v1 = [
            self.row_class.from_dict(
                {"id": data.sha1_2, **self.example_data[0], "tool": tool}
            )
        ]

        # then
        assert actual_data == expected_data_v1

        # given
        data_v2 = data_v1.copy()
        data_v2.update(self.example_data[1])

        endpoint(storage, etype, "add")([self.row_class.from_dict(data_v2)])
        assert summary == expected_summary(1, etype)  # modified so counted

        actual_data = list(endpoint(storage, etype, "get")([data.sha1_2]))

        expected_data_v2 = [
            self.row_class.from_dict(
                {
                    "id": data.sha1_2,
                    **self.example_data[1],
                    "tool": tool,
                }
            )
        ]

        # data did change as the v2 was used to overwrite v1
        assert actual_data == expected_data_v2

    def test_add_deadlock(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        hashes = [
            hash_to_bytes("34973274ccef6ab4dfaaf86599792fa9c3fe4{:03d}".format(i))
            for i in range(1000)
        ]

        data_v1 = [
            self.row_class.from_dict(
                {
                    "id": hash_,
                    **self.example_data[0],
                    "indexer_configuration_id": tool["id"],
                }
            )
            for hash_ in hashes
        ]
        data_v2 = [
            self.row_class.from_dict(
                {
                    "id": hash_,
                    **self.example_data[1],
                    "indexer_configuration_id": tool["id"],
                }
            )
            for hash_ in hashes
        ]

        # Remove one item from each, so that both queries have to succeed for
        # all items to be in the DB.
        data_v2a = data_v2[1:]
        data_v2b = list(reversed(data_v2[0:-1]))

        # given
        endpoint(storage, etype, "add")(data_v1)

        # when
        actual_data = sorted(
            endpoint(storage, etype, "get")(hashes),
            key=lambda x: x.id,
        )

        expected_data_v1 = [
            self.row_class.from_dict(
                {"id": hash_, **self.example_data[0], "tool": tool}
            )
            for hash_ in hashes
        ]

        # then
        assert actual_data == expected_data_v1

        # given
        def f1() -> None:
            endpoint(storage, etype, "add")(data_v2a)

        def f2() -> None:
            endpoint(storage, etype, "add")(data_v2b)

        t1 = threading.Thread(target=f1)
        t2 = threading.Thread(target=f2)
        t2.start()
        t1.start()

        t1.join()
        t2.join()

        actual_data = sorted(
            endpoint(storage, etype, "get")(hashes),
            key=lambda x: x.id,
        )

        expected_data_v2 = [
            self.row_class.from_dict(
                {"id": hash_, **self.example_data[1], "tool": tool}
            )
            for hash_ in hashes
        ]

        assert len(actual_data) == len(expected_data_v1) == len(expected_data_v2)
        for item, expected_item_v1, expected_item_v2 in zip(
            actual_data, expected_data_v1, expected_data_v2
        ):
            assert item in (expected_item_v1, expected_item_v2)

    def test_add__duplicate_twice(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        data_dir1 = self.row_class.from_dict(
            {
                "id": data.directory_id_2,
                **self.example_data[0],
                "indexer_configuration_id": tool["id"],
            }
        )

        data_dir2 = self.row_class.from_dict(
            {
                "id": data.directory_id_2,
                **self.example_data[1],
                "indexer_configuration_id": tool["id"],
            }
        )

        # when
        summary = endpoint(storage, etype, "add")([data_dir1])
        assert summary == expected_summary(1, etype)

        with pytest.raises(DuplicateId):
            endpoint(storage, etype, "add")([data_dir2, data_dir2])

        # then
        actual_data = list(
            endpoint(storage, etype, "get")([data.directory_id_2, data.directory_id_1])
        )

        expected_data = [
            self.row_class.from_dict(
                {"id": data.directory_id_2, **self.example_data[0], "tool": tool}
            )
        ]
        assert actual_data == expected_data

    def test_add(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        # conftest fills it with mimetypes
        storage.journal_writer.journal.objects = []  # type: ignore

        query = [data.sha1_2, data.sha1_1]
        data1 = self.row_class.from_dict(
            {
                "id": data.sha1_2,
                **self.example_data[0],
                "indexer_configuration_id": tool["id"],
            }
        )

        # when
        summary = endpoint(storage, etype, "add")([data1])
        assert summary == expected_summary(1, etype)

        # then
        actual_data = list(endpoint(storage, etype, "get")(query))

        # then
        expected_data = [
            self.row_class.from_dict(
                {"id": data.sha1_2, **self.example_data[0], "tool": tool}
            )
        ]

        assert actual_data == expected_data

        expected_journal_data = _remove_tool_ids(expected_data)

        journal_objects = storage.journal_writer.journal.objects  # type: ignore
        actual_journal_data = [
            obj for (obj_type, obj) in journal_objects if obj_type == self.endpoint_type
        ]
        assert list(sorted(actual_journal_data)) == list(sorted(expected_journal_data))


class TestIndexerStorageContentMimetypes(StorageETypeTester):
    """Test Indexer Storage content_mimetype related methods"""

    endpoint_type = "content_mimetype"
    tool_name = "file"
    example_data = [
        {
            "mimetype": "text/plain",
            "encoding": "utf-8",
        },
        {
            "mimetype": "text/html",
            "encoding": "us-ascii",
        },
    ]
    row_class = ContentMimetypeRow

    def test_generate_content_mimetype_get_partition_failure(
        self, swh_indexer_storage: IndexerStorageInterface
    ) -> None:
        """get_partition call with wrong limit input should fail"""
        storage = swh_indexer_storage
        indexer_configuration_id = 42
        with pytest.raises(
            IndexerStorageArgumentException, match="limit should not be None"
        ):
            storage.content_mimetype_get_partition(
                indexer_configuration_id, 0, 3, limit=None  # type: ignore
            )

    def test_generate_content_mimetype_get_partition_no_limit(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        """get_partition should return result"""
        storage, data = swh_indexer_storage_with_data
        mimetypes = data.mimetypes

        expected_ids = set([c.id for c in mimetypes])
        indexer_configuration_id = mimetypes[0].indexer_configuration_id

        assert len(mimetypes) == 16
        nb_partitions = 16

        actual_ids = []
        for partition_id in range(nb_partitions):
            actual_result = storage.content_mimetype_get_partition(
                indexer_configuration_id, partition_id, nb_partitions
            )
            assert actual_result.next_page_token is None
            actual_ids.extend(actual_result.results)

        assert len(actual_ids) == len(expected_ids)
        for actual_id in actual_ids:
            assert actual_id in expected_ids

    def test_generate_content_mimetype_get_partition_full(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        """get_partition for a single partition should return available ids"""
        storage, data = swh_indexer_storage_with_data
        mimetypes = data.mimetypes
        expected_ids = set([c.id for c in mimetypes])
        indexer_configuration_id = mimetypes[0].indexer_configuration_id

        actual_result = storage.content_mimetype_get_partition(
            indexer_configuration_id, 0, 1
        )
        assert actual_result.next_page_token is None
        actual_ids = actual_result.results
        assert len(actual_ids) == len(expected_ids)
        for actual_id in actual_ids:
            assert actual_id in expected_ids

    def test_generate_content_mimetype_get_partition_empty(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        """get_partition when at least one of the partitions is empty"""
        storage, data = swh_indexer_storage_with_data
        mimetypes = data.mimetypes
        expected_ids = set([c.id for c in mimetypes])
        indexer_configuration_id = mimetypes[0].indexer_configuration_id

        # nb_partitions = smallest power of 2 such that at least one of
        # the partitions is empty
        nb_mimetypes = len(mimetypes)
        nb_partitions = 1 << math.floor(math.log2(nb_mimetypes) + 1)

        seen_ids = []

        for partition_id in range(nb_partitions):
            actual_result = storage.content_mimetype_get_partition(
                indexer_configuration_id,
                partition_id,
                nb_partitions,
                limit=nb_mimetypes + 1,
            )

            for actual_id in actual_result.results:
                seen_ids.append(actual_id)

            # Limit is higher than the max number of results
            assert actual_result.next_page_token is None

        assert set(seen_ids) == expected_ids

    def test_generate_content_mimetype_get_partition_with_pagination(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        """get_partition should return ids provided with pagination"""
        storage, data = swh_indexer_storage_with_data
        mimetypes = data.mimetypes
        expected_ids = set([c.id for c in mimetypes])
        indexer_configuration_id = mimetypes[0].indexer_configuration_id

        nb_partitions = 4

        actual_ids = []
        for partition_id in range(nb_partitions):
            next_page_token = None
            while True:
                actual_result = storage.content_mimetype_get_partition(
                    indexer_configuration_id,
                    partition_id,
                    nb_partitions,
                    limit=2,
                    page_token=next_page_token,
                )
                actual_ids.extend(actual_result.results)
                next_page_token = actual_result.next_page_token
                if next_page_token is None:
                    break

        assert len(set(actual_ids)) == len(set(expected_ids))
        for actual_id in actual_ids:
            assert actual_id in expected_ids


class TestIndexerStorageContentMetadata(StorageETypeTester):
    """Test Indexer Storage content_metadata related methods"""

    tool_name = "swh-metadata-detector"
    endpoint_type = "content_metadata"
    example_data = [
        {
            "metadata": {
                "other": {},
                "codeRepository": {
                    "type": "git",
                    "url": "https://github.com/moranegg/metadata_test",
                },
                "description": "Simple package.json test for indexer",
                "name": "test_metadata",
                "version": "0.0.1",
            },
        },
        {
            "metadata": {"other": {}, "name": "test_metadata", "version": "0.0.1"},
        },
    ]
    row_class = ContentMetadataRow

    def test_add_with_null(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        # conftest fills it with mimetypes
        storage.journal_writer.journal.objects = []  # type: ignore

        query = [data.sha1_2, data.sha1_1]
        data1 = self.row_class.from_dict(
            {
                "id": data.sha1_2,
                "metadata": {"description": "with\u0000nul"},
                "indexer_configuration_id": tool["id"],
            }
        )

        # when
        summary = endpoint(storage, etype, "add")([data1])
        assert summary == expected_summary(1, etype)

        # then
        actual_data = list(endpoint(storage, etype, "get")(query))

        # then
        expected_data_postgresql = [
            self.row_class.from_dict(
                {
                    "id": data.sha1_2,
                    "metadata": {"description": "withnul"},
                    "tool": tool,
                }
            )
        ]
        expected_data_verbatim = [
            self.row_class.from_dict(
                {
                    "id": data.sha1_2,
                    "metadata": {"description": "with\u0000nul"},
                    "tool": tool,
                }
            )
        ]

        assert actual_data in (expected_data_postgresql, expected_data_verbatim)

        expected_journal_data = _remove_tool_ids(expected_data_verbatim)

        journal_objects = storage.journal_writer.journal.objects  # type: ignore
        actual_journal_data = [
            obj for (obj_type, obj) in journal_objects if obj_type == self.endpoint_type
        ]
        assert list(sorted(actual_journal_data)) == list(sorted(expected_journal_data))


class TestIndexerStorageDirectoryIntrinsicMetadata(StorageETypeTester):
    """Test Indexer Storage directory_intrinsic_metadata related methods"""

    tool_name = "swh-metadata-detector"
    endpoint_type = "directory_intrinsic_metadata"
    example_data = [
        {
            "metadata": {
                "other": {},
                "codeRepository": {
                    "type": "git",
                    "url": "https://github.com/moranegg/metadata_test",
                },
                "description": "Simple package.json test for indexer",
                "name": "test_metadata",
                "version": "0.0.1",
            },
            "mappings": ["mapping1"],
        },
        {
            "metadata": {"other": {}, "name": "test_metadata", "version": "0.0.1"},
            "mappings": ["mapping2"],
        },
    ]
    row_class = DirectoryIntrinsicMetadataRow


class TestIndexerStorageContentFossologyLicense(StorageETypeTester):
    endpoint_type = "content_fossology_license"
    tool_name = "nomos"
    example_data = [
        {"license": "Apache-2.0"},
        {"license": "BSD-2-Clause"},
    ]

    row_class = ContentLicenseRow

    # the following tests are disabled because licenses behaves differently
    @pytest.mark.skip
    def test_add__update_in_place_duplicate(self):
        pass

    @pytest.mark.skip
    def test_add_deadlock(self):
        pass

    # content_fossology_license_missing does not exist
    @pytest.mark.skip
    def test_missing(self):
        pass

    def test_content_fossology_license_add__new_license_added(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool = data.tools["nomos"]
        tool_id = tool["id"]

        license1 = ContentLicenseRow(
            id=data.sha1_1,
            license="Apache-2.0",
            indexer_configuration_id=tool_id,
        )

        # given
        storage.content_fossology_license_add([license1])
        # conflict does nothing
        storage.content_fossology_license_add([license1])

        # when
        actual_licenses = list(storage.content_fossology_license_get([data.sha1_1]))

        # then
        expected_licenses = [
            ContentLicenseRow(
                id=data.sha1_1,
                license="Apache-2.0",
                tool=tool,
            )
        ]
        assert actual_licenses == expected_licenses

        # given
        license2 = ContentLicenseRow(
            id=data.sha1_1,
            license="BSD-2-Clause",
            indexer_configuration_id=tool_id,
        )

        storage.content_fossology_license_add([license2])

        actual_licenses = list(storage.content_fossology_license_get([data.sha1_1]))

        expected_licenses.append(
            ContentLicenseRow(
                id=data.sha1_1,
                license="BSD-2-Clause",
                tool=tool,
            )
        )

        # first license was not removed when the second one was added
        assert sorted(actual_licenses) == sorted(expected_licenses)

    def test_generate_content_fossology_license_get_partition_failure(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        """get_partition call with wrong limit input should fail"""
        storage, data = swh_indexer_storage_with_data
        indexer_configuration_id = 42
        with pytest.raises(
            IndexerStorageArgumentException, match="limit should not be None"
        ):
            storage.content_fossology_license_get_partition(
                indexer_configuration_id,
                0,
                3,
                limit=None,  # type: ignore
            )

    def test_generate_content_fossology_license_get_partition_no_limit(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        """get_partition should return results"""
        storage, data = swh_indexer_storage_with_data
        # craft some consistent mimetypes
        fossology_licenses = data.fossology_licenses
        mimetypes = prepare_mimetypes_from_licenses(fossology_licenses)
        indexer_configuration_id = fossology_licenses[0].indexer_configuration_id

        storage.content_mimetype_add(mimetypes)
        # add fossology_licenses to storage
        storage.content_fossology_license_add(fossology_licenses)

        # All ids from the db
        expected_ids = set([c.id for c in fossology_licenses])

        assert len(fossology_licenses) == 10
        assert len(mimetypes) == 10
        nb_partitions = 4

        actual_ids = []
        for partition_id in range(nb_partitions):
            actual_result = storage.content_fossology_license_get_partition(
                indexer_configuration_id, partition_id, nb_partitions
            )
            assert actual_result.next_page_token is None
            actual_ids.extend(actual_result.results)

        assert len(set(actual_ids)) == len(expected_ids)
        for actual_id in actual_ids:
            assert actual_id in expected_ids

    def test_generate_content_fossology_license_get_partition_full(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        """get_partition for a single partition should return available ids"""
        storage, data = swh_indexer_storage_with_data
        # craft some consistent mimetypes
        fossology_licenses = data.fossology_licenses
        mimetypes = prepare_mimetypes_from_licenses(fossology_licenses)
        indexer_configuration_id = fossology_licenses[0].indexer_configuration_id

        storage.content_mimetype_add(mimetypes)
        # add fossology_licenses to storage
        storage.content_fossology_license_add(fossology_licenses)

        # All ids from the db
        expected_ids = set([c.id for c in fossology_licenses])

        actual_result = storage.content_fossology_license_get_partition(
            indexer_configuration_id, 0, 1
        )
        assert actual_result.next_page_token is None
        actual_ids = actual_result.results
        assert len(set(actual_ids)) == len(expected_ids)
        for actual_id in actual_ids:
            assert actual_id in expected_ids

    def test_generate_content_fossology_license_get_partition_empty(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        """get_partition when at least one of the partitions is empty"""
        storage, data = swh_indexer_storage_with_data
        # craft some consistent mimetypes
        fossology_licenses = data.fossology_licenses
        mimetypes = prepare_mimetypes_from_licenses(fossology_licenses)
        indexer_configuration_id = fossology_licenses[0].indexer_configuration_id

        storage.content_mimetype_add(mimetypes)
        # add fossology_licenses to storage
        storage.content_fossology_license_add(fossology_licenses)

        # All ids from the db
        expected_ids = set([c.id for c in fossology_licenses])

        # nb_partitions = smallest power of 2 such that at least one of
        # the partitions is empty
        nb_licenses = len(fossology_licenses)
        nb_partitions = 1 << math.floor(math.log2(nb_licenses) + 1)

        seen_ids = []

        for partition_id in range(nb_partitions):
            actual_result = storage.content_fossology_license_get_partition(
                indexer_configuration_id,
                partition_id,
                nb_partitions,
                limit=nb_licenses + 1,
            )

            for actual_id in actual_result.results:
                seen_ids.append(actual_id)

            # Limit is higher than the max number of results
            assert actual_result.next_page_token is None

        assert set(seen_ids) == expected_ids

    def test_generate_content_fossology_license_get_partition_with_pagination(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        """get_partition should return ids provided with paginationv"""
        storage, data = swh_indexer_storage_with_data
        # craft some consistent mimetypes
        fossology_licenses = data.fossology_licenses
        mimetypes = prepare_mimetypes_from_licenses(fossology_licenses)
        indexer_configuration_id = fossology_licenses[0].indexer_configuration_id

        storage.content_mimetype_add(mimetypes)
        # add fossology_licenses to storage
        storage.content_fossology_license_add(fossology_licenses)

        # All ids from the db
        expected_ids = [c.id for c in fossology_licenses]

        nb_partitions = 4

        actual_ids = []
        for partition_id in range(nb_partitions):
            next_page_token = None
            while True:
                actual_result = storage.content_fossology_license_get_partition(
                    indexer_configuration_id,
                    partition_id,
                    nb_partitions,
                    limit=2,
                    page_token=next_page_token,
                )
                actual_ids.extend(actual_result.results)
                next_page_token = actual_result.next_page_token
                if next_page_token is None:
                    break

        assert len(set(actual_ids)) == len(set(expected_ids))
        for actual_id in actual_ids:
            assert actual_id in expected_ids

    def test_add_empty(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        (storage, data) = swh_indexer_storage_with_data
        etype = self.endpoint_type

        summary = endpoint(storage, etype, "add")([])
        assert summary == {"content_fossology_license:add": 0}

        actual_license = list(endpoint(storage, etype, "get")([data.sha1_2]))

        assert actual_license == []

    def test_get_unknown(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        (storage, data) = swh_indexer_storage_with_data
        etype = self.endpoint_type

        actual_license = list(endpoint(storage, etype, "get")([data.sha1_2]))

        assert actual_license == []


class TestIndexerStorageOriginIntrinsicMetadata:
    def test_origin_intrinsic_metadata_add(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        metadata = {
            "version": None,
            "name": None,
        }
        metadata_dir = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_2,
            metadata=metadata,
            mappings=["mapping1"],
            indexer_configuration_id=tool_id,
        )
        metadata_origin = OriginIntrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata,
            indexer_configuration_id=tool_id,
            mappings=["mapping1"],
            from_directory=data.directory_id_2,
        )

        # when
        storage.directory_intrinsic_metadata_add([metadata_dir])
        storage.origin_intrinsic_metadata_add([metadata_origin])

        # then
        actual_metadata = list(
            storage.origin_intrinsic_metadata_get([data.origin_url_1, "no://where"])
        )

        expected_metadata = [
            OriginIntrinsicMetadataRow(
                id=data.origin_url_1,
                metadata=metadata,
                tool=data.tools["swh-metadata-detector"],
                from_directory=data.directory_id_2,
                mappings=["mapping1"],
            )
        ]

        assert actual_metadata == expected_metadata

        expected_journal_metadata = _remove_tool_ids(expected_metadata)

        journal_objects = storage.journal_writer.journal.objects  # type: ignore
        actual_journal_metadata = [
            obj
            for (obj_type, obj) in journal_objects
            if obj_type == "origin_intrinsic_metadata"
        ]
        assert list(sorted(actual_journal_metadata)) == list(
            sorted(expected_journal_metadata)
        )

    def test_origin_intrinsic_metadata_add_update_in_place_duplicate(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        metadata_v1: Dict[str, Any] = {
            "version": None,
            "name": None,
        }
        metadata_dir_v1 = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_2,
            metadata=metadata_v1,
            mappings=[],
            indexer_configuration_id=tool_id,
        )
        metadata_origin_v1 = OriginIntrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata_v1.copy(),
            indexer_configuration_id=tool_id,
            mappings=[],
            from_directory=data.directory_id_2,
        )

        # given
        storage.directory_intrinsic_metadata_add([metadata_dir_v1])
        storage.origin_intrinsic_metadata_add([metadata_origin_v1])

        # when
        actual_metadata = list(
            storage.origin_intrinsic_metadata_get([data.origin_url_1])
        )

        # then
        expected_metadata_v1 = [
            OriginIntrinsicMetadataRow(
                id=data.origin_url_1,
                metadata=metadata_v1,
                tool=data.tools["swh-metadata-detector"],
                from_directory=data.directory_id_2,
                mappings=[],
            )
        ]
        assert actual_metadata == expected_metadata_v1

        # given
        metadata_v2 = metadata_v1.copy()
        metadata_v2.update(
            {
                "name": "test_update_duplicated_metadata",
                "author": "MG",
            }
        )
        metadata_dir_v2 = attr.evolve(metadata_dir_v1, metadata=metadata_v2)
        metadata_origin_v2 = OriginIntrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata_v2.copy(),
            indexer_configuration_id=tool_id,
            mappings=["npm"],
            from_directory=data.directory_id_1,
        )

        storage.directory_intrinsic_metadata_add([metadata_dir_v2])
        storage.origin_intrinsic_metadata_add([metadata_origin_v2])

        actual_metadata = list(
            storage.origin_intrinsic_metadata_get([data.origin_url_1])
        )

        expected_metadata_v2 = [
            OriginIntrinsicMetadataRow(
                id=data.origin_url_1,
                metadata=metadata_v2,
                tool=data.tools["swh-metadata-detector"],
                from_directory=data.directory_id_1,
                mappings=["npm"],
            )
        ]

        # metadata did change as the v2 was used to overwrite v1
        assert actual_metadata == expected_metadata_v2

    def test_origin_intrinsic_metadata_add__deadlock(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        origins = ["file:///tmp/origin{:02d}".format(i) for i in range(100)]

        example_data1: Dict[str, Any] = {
            "metadata": {
                "version": None,
                "name": None,
            },
            "mappings": [],
        }
        example_data2: Dict[str, Any] = {
            "metadata": {
                "version": "v1.1.1",
                "name": "foo",
            },
            "mappings": [],
        }

        metadata_dir_v1 = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_2,
            metadata={
                "version": None,
                "name": None,
            },
            mappings=[],
            indexer_configuration_id=tool_id,
        )

        data_v1 = [
            OriginIntrinsicMetadataRow(
                id=origin,
                from_directory=data.directory_id_2,
                indexer_configuration_id=tool_id,
                **example_data1,
            )
            for origin in origins
        ]
        data_v2 = [
            OriginIntrinsicMetadataRow(
                id=origin,
                from_directory=data.directory_id_2,
                indexer_configuration_id=tool_id,
                **example_data2,
            )
            for origin in origins
        ]

        # Remove one item from each, so that both queries have to succeed for
        # all items to be in the DB.
        data_v2a = data_v2[1:]
        data_v2b = list(reversed(data_v2[0:-1]))

        # given
        storage.directory_intrinsic_metadata_add([metadata_dir_v1])
        storage.origin_intrinsic_metadata_add(data_v1)

        # when
        actual_data = list(storage.origin_intrinsic_metadata_get(origins))

        expected_data_v1 = [
            OriginIntrinsicMetadataRow(
                id=origin,
                from_directory=data.directory_id_2,
                tool=data.tools["swh-metadata-detector"],
                **example_data1,
            )
            for origin in origins
        ]

        # then
        assert actual_data == expected_data_v1

        # given
        def f1() -> None:
            storage.origin_intrinsic_metadata_add(data_v2a)

        def f2() -> None:
            storage.origin_intrinsic_metadata_add(data_v2b)

        t1 = threading.Thread(target=f1)
        t2 = threading.Thread(target=f2)
        t2.start()
        t1.start()

        t1.join()
        t2.join()

        actual_data = list(storage.origin_intrinsic_metadata_get(origins))

        expected_data_v2 = [
            OriginIntrinsicMetadataRow(
                id=origin,
                from_directory=data.directory_id_2,
                tool=data.tools["swh-metadata-detector"],
                **example_data2,
            )
            for origin in origins
        ]

        actual_data.sort(key=lambda item: item.id)
        assert len(actual_data) == len(expected_data_v1) == len(expected_data_v2)
        for item, expected_item_v1, expected_item_v2 in zip(
            actual_data, expected_data_v1, expected_data_v2
        ):
            assert item in (expected_item_v1, expected_item_v2)

    def test_origin_intrinsic_metadata_add__duplicate_twice(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        metadata = {
            "developmentStatus": None,
            "name": None,
        }
        metadata_dir = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_2,
            metadata=metadata,
            mappings=["mapping1"],
            indexer_configuration_id=tool_id,
        )
        metadata_origin = OriginIntrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata,
            indexer_configuration_id=tool_id,
            mappings=["mapping1"],
            from_directory=data.directory_id_2,
        )

        # when
        storage.directory_intrinsic_metadata_add([metadata_dir])

        with pytest.raises(DuplicateId):
            storage.origin_intrinsic_metadata_add([metadata_origin, metadata_origin])

    def test_origin_intrinsic_metadata_search_fulltext(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        metadata1 = {
            "author": "John Doe",
        }
        metadata1_dir = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_1,
            metadata=metadata1,
            mappings=[],
            indexer_configuration_id=tool_id,
        )
        metadata1_origin = OriginIntrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata1,
            mappings=[],
            indexer_configuration_id=tool_id,
            from_directory=data.directory_id_1,
        )
        metadata2 = {
            "author": "Jane Doe",
        }
        metadata2_dir = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_2,
            metadata=metadata2,
            mappings=[],
            indexer_configuration_id=tool_id,
        )
        metadata2_origin = OriginIntrinsicMetadataRow(
            id=data.origin_url_2,
            metadata=metadata2,
            mappings=[],
            indexer_configuration_id=tool_id,
            from_directory=data.directory_id_2,
        )

        # when
        storage.directory_intrinsic_metadata_add([metadata1_dir])
        storage.origin_intrinsic_metadata_add([metadata1_origin])
        storage.directory_intrinsic_metadata_add([metadata2_dir])
        storage.origin_intrinsic_metadata_add([metadata2_origin])

        # then
        search = storage.origin_intrinsic_metadata_search_fulltext
        assert set([res.id for res in search(["Doe"])]) == set(
            [data.origin_url_1, data.origin_url_2]
        )
        assert [res.id for res in search(["John", "Doe"])] == [data.origin_url_1]
        assert [res.id for res in search(["John"])] == [data.origin_url_1]
        assert not list(search(["John", "Jane"]))

    def test_origin_intrinsic_metadata_search_fulltext_rank(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        # The following authors have "Random Person" to add some more content
        # to the JSON data, to work around normalization quirks when there
        # are few words (rank/(1+ln(nb_words)) is very sensitive to nb_words
        # for small values of nb_words).
        metadata1 = {
            "author": [
                "Random Person",
                "John Doe",
                "Jane Doe",
            ]
        }
        metadata1_dir = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_1,
            metadata=metadata1,
            mappings=[],
            indexer_configuration_id=tool_id,
        )
        metadata1_origin = OriginIntrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata1,
            mappings=[],
            indexer_configuration_id=tool_id,
            from_directory=data.directory_id_1,
        )
        metadata2 = {
            "author": [
                "Random Person",
                "Jane Doe",
            ]
        }
        metadata2_dir = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_2,
            metadata=metadata2,
            mappings=[],
            indexer_configuration_id=tool_id,
        )
        metadata2_origin = OriginIntrinsicMetadataRow(
            id=data.origin_url_2,
            metadata=metadata2,
            mappings=[],
            indexer_configuration_id=tool_id,
            from_directory=data.directory_id_2,
        )

        # when
        storage.directory_intrinsic_metadata_add([metadata1_dir])
        storage.origin_intrinsic_metadata_add([metadata1_origin])
        storage.directory_intrinsic_metadata_add([metadata2_dir])
        storage.origin_intrinsic_metadata_add([metadata2_origin])

        # then
        search = storage.origin_intrinsic_metadata_search_fulltext
        assert [res.id for res in search(["Doe"])] == [
            data.origin_url_1,
            data.origin_url_2,
        ]
        assert [res.id for res in search(["Doe"], limit=1)] == [data.origin_url_1]
        assert [res.id for res in search(["John"])] == [data.origin_url_1]
        assert [res.id for res in search(["Jane"])] == [
            data.origin_url_2,
            data.origin_url_1,
        ]
        assert [res.id for res in search(["John", "Jane"])] == [data.origin_url_1]

    def _fill_origin_intrinsic_metadata(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        tool1_id = data.tools["swh-metadata-detector"]["id"]
        tool2_id = data.tools["swh-metadata-detector2"]["id"]

        metadata1 = {
            "@context": "foo",
            "author": "John Doe",
        }
        metadata1_dir = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_1,
            metadata=metadata1,
            mappings=["npm"],
            indexer_configuration_id=tool1_id,
        )
        metadata1_origin = OriginIntrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata1,
            mappings=["npm"],
            indexer_configuration_id=tool1_id,
            from_directory=data.directory_id_1,
        )
        metadata2 = {
            "@context": "foo",
            "author": "Jane Doe",
        }
        metadata2_dir = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_2,
            metadata=metadata2,
            mappings=["npm", "gemspec"],
            indexer_configuration_id=tool2_id,
        )
        metadata2_origin = OriginIntrinsicMetadataRow(
            id=data.origin_url_2,
            metadata=metadata2,
            mappings=["npm", "gemspec"],
            indexer_configuration_id=tool2_id,
            from_directory=data.directory_id_2,
        )
        metadata3 = {
            "@context": "foo",
        }
        metadata3_dir = DirectoryIntrinsicMetadataRow(
            id=data.directory_id_3,
            metadata=metadata3,
            mappings=["npm", "gemspec"],
            indexer_configuration_id=tool2_id,
        )
        metadata3_origin = OriginIntrinsicMetadataRow(
            id=data.origin_url_3,
            metadata=metadata3,
            mappings=["pkg-info"],
            indexer_configuration_id=tool2_id,
            from_directory=data.directory_id_3,
        )

        storage.directory_intrinsic_metadata_add([metadata1_dir])
        storage.origin_intrinsic_metadata_add([metadata1_origin])
        storage.directory_intrinsic_metadata_add([metadata2_dir])
        storage.origin_intrinsic_metadata_add([metadata2_origin])
        storage.directory_intrinsic_metadata_add([metadata3_dir])
        storage.origin_intrinsic_metadata_add([metadata3_origin])

    def test_origin_intrinsic_metadata_search_by_producer(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        self._fill_origin_intrinsic_metadata(swh_indexer_storage_with_data)
        tool1 = data.tools["swh-metadata-detector"]
        tool2 = data.tools["swh-metadata-detector2"]
        endpoint = storage.origin_intrinsic_metadata_search_by_producer

        # test pagination
        # no 'page_token' param, return all origins
        result = endpoint(ids_only=True)
        assert result == PagedResult(
            results=[
                data.origin_url_1,
                data.origin_url_2,
                data.origin_url_3,
            ],
            next_page_token=None,
        )

        # 'page_token' is < than origin_1, return everything
        result = endpoint(page_token=data.origin_url_1[:-1], ids_only=True)
        assert result == PagedResult(
            results=[
                data.origin_url_1,
                data.origin_url_2,
                data.origin_url_3,
            ],
            next_page_token=None,
        )

        # 'page_token' is origin_3, return nothing
        result = endpoint(page_token=data.origin_url_3, ids_only=True)
        assert result == PagedResult(results=[], next_page_token=None)

        # test limit argument
        result = endpoint(page_token=data.origin_url_1[:-1], limit=2, ids_only=True)
        assert result == PagedResult(
            results=[data.origin_url_1, data.origin_url_2],
            next_page_token=data.origin_url_2,
        )

        result = endpoint(page_token=data.origin_url_1, limit=2, ids_only=True)
        assert result == PagedResult(
            results=[data.origin_url_2, data.origin_url_3],
            next_page_token=None,
        )

        result = endpoint(page_token=data.origin_url_2, limit=2, ids_only=True)
        assert result == PagedResult(
            results=[data.origin_url_3],
            next_page_token=None,
        )

        # test mappings filtering
        result = endpoint(mappings=["npm"], ids_only=True)
        assert result == PagedResult(
            results=[data.origin_url_1, data.origin_url_2],
            next_page_token=None,
        )

        result = endpoint(mappings=["npm", "gemspec"], ids_only=True)
        assert result == PagedResult(
            results=[data.origin_url_1, data.origin_url_2],
            next_page_token=None,
        )

        result = endpoint(mappings=["gemspec"], ids_only=True)
        assert result == PagedResult(
            results=[data.origin_url_2],
            next_page_token=None,
        )

        result = endpoint(mappings=["pkg-info"], ids_only=True)
        assert result == PagedResult(
            results=[data.origin_url_3],
            next_page_token=None,
        )

        result = endpoint(mappings=["foobar"], ids_only=True)
        assert result == PagedResult(
            results=[],
            next_page_token=None,
        )

        # test pagination + mappings
        result = endpoint(mappings=["npm"], limit=1, ids_only=True)
        assert result == PagedResult(
            results=[data.origin_url_1],
            next_page_token=data.origin_url_1,
        )

        # test tool filtering
        result = endpoint(tool_ids=[tool1["id"]], ids_only=True)
        assert result == PagedResult(
            results=[data.origin_url_1],
            next_page_token=None,
        )

        result = endpoint(tool_ids=[tool2["id"]], ids_only=True)
        assert sorted(result.results) == [data.origin_url_2, data.origin_url_3]
        assert result.next_page_token is None

        result = endpoint(tool_ids=[tool1["id"], tool2["id"]], ids_only=True)
        assert sorted(result.results) == [
            data.origin_url_1,
            data.origin_url_2,
            data.origin_url_3,
        ]
        assert result.next_page_token is None

        # test ids_only=False
        assert endpoint(mappings=["gemspec"]) == PagedResult(
            results=[
                OriginIntrinsicMetadataRow(
                    id=data.origin_url_2,
                    metadata={
                        "@context": "foo",
                        "author": "Jane Doe",
                    },
                    mappings=["npm", "gemspec"],
                    tool=tool2,
                    from_directory=data.directory_id_2,
                )
            ],
            next_page_token=None,
        )

    def test_origin_intrinsic_metadata_stats(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        self._fill_origin_intrinsic_metadata(swh_indexer_storage_with_data)

        result = storage.origin_intrinsic_metadata_stats()
        assert result == {
            "per_mapping": {
                "cff": 0,
                "gemspec": 1,
                "npm": 2,
                "pkg-info": 1,
                "codemeta": 0,
                "maven": 0,
            },
            "total": 3,
            "non_empty": 2,
        }


class TestIndexerStorageOriginExtrinsicMetadata:
    def test_origin_extrinsic_metadata_add(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        metadata = {
            "version": None,
            "name": None,
        }
        metadata_origin = OriginExtrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata,
            indexer_configuration_id=tool_id,
            mappings=["mapping1"],
            from_remd_id=b"\x02" * 20,
        )

        # when
        storage.origin_extrinsic_metadata_add([metadata_origin])

        # then
        actual_metadata = list(
            storage.origin_extrinsic_metadata_get([data.origin_url_1, "no://where"])
        )

        expected_metadata = [
            OriginExtrinsicMetadataRow(
                id=data.origin_url_1,
                metadata=metadata,
                tool=data.tools["swh-metadata-detector"],
                from_remd_id=b"\x02" * 20,
                mappings=["mapping1"],
            )
        ]

        assert actual_metadata == expected_metadata

        expected_journal_metadata = _remove_tool_ids(expected_metadata)

        journal_objects = storage.journal_writer.journal.objects  # type: ignore
        actual_journal_metadata = [
            obj
            for (obj_type, obj) in journal_objects
            if obj_type == "origin_extrinsic_metadata"
        ]
        assert list(sorted(actual_journal_metadata)) == list(
            sorted(expected_journal_metadata)
        )

    def test_origin_extrinsic_metadata_add_update_in_place_duplicate(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        metadata_v1: Dict[str, Any] = {
            "version": None,
            "name": None,
        }
        metadata_origin_v1 = OriginExtrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata_v1.copy(),
            indexer_configuration_id=tool_id,
            mappings=[],
            from_remd_id=b"\x02" * 20,
        )

        # given
        storage.origin_extrinsic_metadata_add([metadata_origin_v1])

        # when
        actual_metadata = list(
            storage.origin_extrinsic_metadata_get([data.origin_url_1])
        )

        # then
        expected_metadata_v1 = [
            OriginExtrinsicMetadataRow(
                id=data.origin_url_1,
                metadata=metadata_v1,
                tool=data.tools["swh-metadata-detector"],
                from_remd_id=b"\x02" * 20,
                mappings=[],
            )
        ]
        assert actual_metadata == expected_metadata_v1

        # given
        metadata_v2 = metadata_v1.copy()
        metadata_v2.update(
            {
                "name": "test_update_duplicated_metadata",
                "author": "MG",
            }
        )
        metadata_origin_v2 = OriginExtrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata_v2.copy(),
            indexer_configuration_id=tool_id,
            mappings=["github"],
            from_remd_id=b"\x02" * 20,
        )

        storage.origin_extrinsic_metadata_add([metadata_origin_v2])

        actual_metadata = list(
            storage.origin_extrinsic_metadata_get([data.origin_url_1])
        )

        expected_metadata_v2 = [
            OriginExtrinsicMetadataRow(
                id=data.origin_url_1,
                metadata=metadata_v2,
                tool=data.tools["swh-metadata-detector"],
                from_remd_id=b"\x02" * 20,
                mappings=["github"],
            )
        ]

        # metadata did change as the v2 was used to overwrite v1
        assert actual_metadata == expected_metadata_v2

    def test_origin_extrinsic_metadata_add__deadlock(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        origins = ["file:///tmp/origin{:02d}".format(i) for i in range(100)]

        example_data1: Dict[str, Any] = {
            "metadata": {
                "version": None,
                "name": None,
            },
            "mappings": [],
        }
        example_data2: Dict[str, Any] = {
            "metadata": {
                "version": "v1.1.1",
                "name": "foo",
            },
            "mappings": [],
        }

        data_v1 = [
            OriginExtrinsicMetadataRow(
                id=origin,
                from_remd_id=b"\x02" * 20,
                indexer_configuration_id=tool_id,
                **example_data1,
            )
            for origin in origins
        ]
        data_v2 = [
            OriginExtrinsicMetadataRow(
                id=origin,
                from_remd_id=b"\x02" * 20,
                indexer_configuration_id=tool_id,
                **example_data2,
            )
            for origin in origins
        ]

        # Remove one item from each, so that both queries have to succeed for
        # all items to be in the DB.
        data_v2a = data_v2[1:]
        data_v2b = list(reversed(data_v2[0:-1]))

        # given
        storage.origin_extrinsic_metadata_add(data_v1)

        # when
        actual_data = list(storage.origin_extrinsic_metadata_get(origins))

        expected_data_v1 = [
            OriginExtrinsicMetadataRow(
                id=origin,
                from_remd_id=b"\x02" * 20,
                tool=data.tools["swh-metadata-detector"],
                **example_data1,
            )
            for origin in origins
        ]

        # then
        assert actual_data == expected_data_v1

        # given
        def f1() -> None:
            storage.origin_extrinsic_metadata_add(data_v2a)

        def f2() -> None:
            storage.origin_extrinsic_metadata_add(data_v2b)

        t1 = threading.Thread(target=f1)
        t2 = threading.Thread(target=f2)
        t2.start()
        t1.start()

        t1.join()
        t2.join()

        actual_data = list(storage.origin_extrinsic_metadata_get(origins))

        expected_data_v2 = [
            OriginExtrinsicMetadataRow(
                id=origin,
                from_remd_id=b"\x02" * 20,
                tool=data.tools["swh-metadata-detector"],
                **example_data2,
            )
            for origin in origins
        ]

        actual_data.sort(key=lambda item: item.id)
        assert len(actual_data) == len(expected_data_v1) == len(expected_data_v2)
        for item, expected_item_v1, expected_item_v2 in zip(
            actual_data, expected_data_v1, expected_data_v2
        ):
            assert item in (expected_item_v1, expected_item_v2)

    def test_origin_extrinsic_metadata_add__duplicate_twice(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools["swh-metadata-detector"]["id"]

        metadata = {
            "developmentStatus": None,
            "name": None,
        }
        metadata_origin = OriginExtrinsicMetadataRow(
            id=data.origin_url_1,
            metadata=metadata,
            indexer_configuration_id=tool_id,
            mappings=["mapping1"],
            from_remd_id=b"\x02" * 20,
        )

        # when
        with pytest.raises(DuplicateId):
            storage.origin_extrinsic_metadata_add([metadata_origin, metadata_origin])


class TestIndexerStorageIndexerConfiguration:
    def test_indexer_configuration_add(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        tool = {
            "tool_name": "some-unknown-tool",
            "tool_version": "some-version",
            "tool_configuration": {"debian-package": "some-package"},
        }

        actual_tool = storage.indexer_configuration_get(tool)
        assert actual_tool is None  # does not exist

        # add it
        actual_tools = list(storage.indexer_configuration_add([tool]))

        assert len(actual_tools) == 1
        actual_tool = actual_tools[0]
        assert actual_tool is not None  # now it exists
        new_id = actual_tool.pop("id")
        assert actual_tool == tool

        actual_tools2 = list(storage.indexer_configuration_add([tool]))
        actual_tool2 = actual_tools2[0]
        assert actual_tool2 is not None  # now it exists
        new_id2 = actual_tool2.pop("id")

        assert new_id == new_id2
        assert actual_tool == actual_tool2

    def test_indexer_configuration_add_multiple(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        tool = {
            "tool_name": "some-unknown-tool",
            "tool_version": "some-version",
            "tool_configuration": {"debian-package": "some-package"},
        }

        actual_tools = list(storage.indexer_configuration_add([tool]))
        assert len(actual_tools) == 1

        new_tools = [
            tool,
            {
                "tool_name": "yet-another-tool",
                "tool_version": "version",
                "tool_configuration": {},
            },
        ]

        actual_tools = list(storage.indexer_configuration_add(new_tools))
        assert len(actual_tools) == 2

        # order not guaranteed, so we iterate over results to check
        for tool in actual_tools:
            _id = tool.pop("id")
            assert _id is not None
            assert tool in new_tools

    def test_indexer_configuration_get_missing(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        tool = {
            "tool_name": "unknown-tool",
            "tool_version": "3.1.0rc2-31-ga2cbb8c",
            "tool_configuration": {"command_line": "nomossa <filepath>"},
        }

        actual_tool = storage.indexer_configuration_get(tool)

        assert actual_tool is None

    def test_indexer_configuration_get(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        tool = {
            "tool_name": "nomos",
            "tool_version": "3.1.0rc2-31-ga2cbb8c",
            "tool_configuration": {"command_line": "nomossa <filepath>"},
        }

        actual_tool = storage.indexer_configuration_get(tool)
        assert actual_tool

        expected_tool = tool.copy()
        del actual_tool["id"]

        assert expected_tool == actual_tool

    def test_indexer_configuration_metadata_get_missing_context(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        tool = {
            "tool_name": "swh-metadata-translator",
            "tool_version": "0.0.1",
            "tool_configuration": {"context": "unknown-context"},
        }

        actual_tool = storage.indexer_configuration_get(tool)

        assert actual_tool is None

    def test_indexer_configuration_metadata_get(
        self, swh_indexer_storage_with_data: Tuple[IndexerStorageInterface, Any]
    ) -> None:
        storage, data = swh_indexer_storage_with_data
        tool = {
            "tool_name": "swh-metadata-translator",
            "tool_version": "0.0.1",
            "tool_configuration": {"type": "local", "context": "NpmMapping"},
        }

        storage.indexer_configuration_add([tool])
        actual_tool = storage.indexer_configuration_get(tool)
        assert actual_tool

        expected_tool = tool.copy()
        expected_tool["id"] = actual_tool["id"]

        assert expected_tool == actual_tool



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import datetime
import functools
from typing import Any, Dict, List, Tuple
import unittest

from hypothesis import strategies

from swh.core.api.classes import stream_results
from swh.indexer.storage import INDEXER_CFG_KEY
from swh.model.hashutil import hash_to_bytes
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    ObjectType,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Person,
    Release,
    Revision,
    RevisionType,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
    TimestampWithTimezone,
)
from swh.storage.utils import now

BASE_TEST_CONFIG: Dict[str, Dict[str, Any]] = {
    "storage": {"cls": "memory"},
    "objstorage": {"cls": "memory"},
    INDEXER_CFG_KEY: {"cls": "memory"},
}

ORIGIN_VISITS = [
    {"type": "git", "origin": "https://github.com/SoftwareHeritage/swh-storage"},
    {"type": "ftp", "origin": "rsync://ftp.gnu.org/gnu/3dldf"},
    {
        "type": "deposit",
        "origin": "https://forge.softwareheritage.org/source/jesuisgpl/",
    },
    {
        "type": "pypi",
        "origin": "https://old-pypi.example.org/project/limnoria/",
    },  # with rev head
    {"type": "pypi", "origin": "https://pypi.org/project/limnoria/"},  # with rel head
    {"type": "svn", "origin": "http://0-512-md.googlecode.com/svn/"},
    {"type": "git", "origin": "https://github.com/librariesio/yarn-parser"},
    {"type": "git", "origin": "https://github.com/librariesio/yarn-parser.git"},
    {"type": "git", "origin": "https://npm.example.org/yarn-parser"},
]

ORIGINS = [Origin(url=visit["origin"]) for visit in ORIGIN_VISITS]

OBJ_STORAGE_RAW_CONTENT: Dict[str, bytes] = {
    "text:some": b"this is some text",
    "text:another": b"another text",
    "text:yet": b"yet another text",
    "python:code": b"""
    import unittest
    import logging
    from swh.indexer.mimetype import MimetypeIndexer
    from swh.indexer.tests.test_utils import MockObjStorage

    class MockStorage():
        def content_mimetype_add(self, mimetypes):
            self.state = mimetypes

        def indexer_configuration_add(self, tools):
            return [{
                'id': 10,
            }]
    """,
    "c:struct": b"""
        #ifndef __AVL__
        #define __AVL__

        typedef struct _avl_tree avl_tree;

        typedef struct _data_t {
          int content;
        } data_t;
    """,
    "lisp:assertion": b"""
    (should 'pygments (recognize 'lisp 'easily))

    """,
    "json:test-metadata-package.json": b"""
    {
        "name": "test_metadata",
        "version": "0.0.1",
        "description": "Simple package.json test for indexer",
        "repository": {
          "type": "git",
          "url": "https://github.com/moranegg/metadata_test"
      }
    }
    """,
    "json:npm-package.json": b"""
    {
      "version": "5.0.3",
      "name": "npm",
      "description": "a package manager for JavaScript",
      "preferGlobal": true,
      "config": {
        "publishtest": false
      },
      "homepage": "https://docs.npmjs.com/",
      "author": "Isaac Z. Schlueter <i@izs.me> (http://blog.izs.me)",
      "repository": {
        "type": "git",
        "url": "https://github.com/npm/npm"
      },
      "bugs": {
        "url": "https://github.com/npm/npm/issues"
      },
      "dependencies": {
        "JSONStream": "~1.3.1",
        "abbrev": "~1.1.0",
        "ansi-regex": "~2.1.1",
        "ansicolors": "~0.3.2",
        "ansistyles": "~0.1.3"
      },
      "devDependencies": {
        "tacks": "~1.2.6",
        "tap": "~10.3.2"
      },
      "license": "Artistic-2.0"
    }

    """,
    "text:carriage-return": b"""
    """,
    "text:empty": b"",
    # was 626364 / b'bcd'
    "text:unimportant": b"unimportant content for bcd",
    # was 636465 / b'cde' now yarn-parser package.json
    "json:yarn-parser-package.json": b"""
    {
      "name": "yarn-parser",
      "version": "1.0.0",
      "description": "Tiny web service for parsing yarn.lock files",
      "main": "index.js",
      "scripts": {
        "start": "node index.js",
        "test": "mocha"
      },
      "engines": {
        "node": "9.8.0"
      },
      "repository": {
        "type": "git",
        "url": "git+https://github.com/librariesio/yarn-parser.git"
      },
      "author": "Andrew Nesbitt",
      "license": "AGPL-3.0",
      "bugs": {
        "url": "https://github.com/librariesio/yarn-parser/issues"
      },
      "homepage": "https://github.com/librariesio/yarn-parser#readme",
      "dependencies": {
        "@yarnpkg/lockfile": "^1.0.0",
        "body-parser": "^1.15.2",
        "express": "^4.14.0"
      },
      "devDependencies": {
        "chai": "^4.1.2",
        "mocha": "^5.2.0",
        "request": "^2.87.0",
        "test": "^0.6.0"
      }
    }

""",
}

MAPPING_DESCRIPTION_CONTENT_SHA1GIT: Dict[str, bytes] = {}
MAPPING_DESCRIPTION_CONTENT_SHA1: Dict[str, bytes] = {}
OBJ_STORAGE_DATA: Dict[bytes, bytes] = {}

for key_description, data in OBJ_STORAGE_RAW_CONTENT.items():
    content = Content.from_data(data)
    MAPPING_DESCRIPTION_CONTENT_SHA1GIT[key_description] = content.sha1_git
    MAPPING_DESCRIPTION_CONTENT_SHA1[key_description] = content.sha1
    OBJ_STORAGE_DATA[content.sha1] = data


RAW_CONTENT_METADATA = [
    (
        "du français".encode(),
        "text/plain",
        "utf-8",
    ),
    (
        b"def __init__(self):",
        ("text/x-python", "text/x-script.python"),
        "us-ascii",
    ),
    (
        b"\xff\xfe\x00\x00\x00\x00\xff\xfe\xff\xff",
        "application/octet-stream",
        "",
    ),
]

RAW_CONTENTS: Dict[bytes, Tuple] = {}
RAW_CONTENT_IDS: List[bytes] = []

for index, raw_content_d in enumerate(RAW_CONTENT_METADATA):
    raw_content = raw_content_d[0]
    content = Content.from_data(raw_content)
    RAW_CONTENTS[content.sha1] = raw_content_d
    RAW_CONTENT_IDS.append(content.sha1)
    # and write it to objstorage data so it's flushed in the objstorage
    OBJ_STORAGE_DATA[content.sha1] = raw_content


SHA1_TO_LICENSES: Dict[bytes, List[str]] = {
    RAW_CONTENT_IDS[0]: ["GPL"],
    RAW_CONTENT_IDS[1]: ["AGPL"],
    RAW_CONTENT_IDS[2]: [],
}


DIRECTORY = Directory(
    entries=(
        DirectoryEntry(
            name=b"index.js",
            type="file",
            target=MAPPING_DESCRIPTION_CONTENT_SHA1GIT["text:some"],
            perms=0o100644,
        ),
        DirectoryEntry(
            name=b"package.json",
            type="file",
            target=MAPPING_DESCRIPTION_CONTENT_SHA1GIT[
                "json:test-metadata-package.json"
            ],
            perms=0o100644,
        ),
        DirectoryEntry(
            name=b".github",
            type="dir",
            target=Directory(entries=()).id,
            perms=0o040000,
        ),
    ),
)

DIRECTORY2 = Directory(
    entries=(
        DirectoryEntry(
            name=b"package.json",
            type="file",
            target=MAPPING_DESCRIPTION_CONTENT_SHA1GIT["json:yarn-parser-package.json"],
            perms=0o100644,
        ),
    ),
)

_utc_plus_2 = datetime.timezone(datetime.timedelta(minutes=120))

REVISION = Revision(
    message=b"Improve search functionality",
    author=Person(
        name=b"Andrew Nesbitt",
        fullname=b"Andrew Nesbitt <andrewnez@gmail.com>",
        email=b"andrewnez@gmail.com",
    ),
    committer=Person(
        name=b"Andrew Nesbitt",
        fullname=b"Andrew Nesbitt <andrewnez@gmail.com>",
        email=b"andrewnez@gmail.com",
    ),
    committer_date=TimestampWithTimezone.from_datetime(
        datetime.datetime(2013, 10, 4, 12, 50, 49, tzinfo=_utc_plus_2)
    ),
    type=RevisionType.GIT,
    synthetic=False,
    date=TimestampWithTimezone.from_datetime(
        datetime.datetime(2017, 2, 20, 16, 14, 16, tzinfo=_utc_plus_2)
    ),
    directory=DIRECTORY2.id,
    parents=(),
)

REVISIONS = [REVISION]

RELEASE = Release(
    name=b"v0.0.0",
    message=None,
    author=Person(
        name=b"Andrew Nesbitt",
        fullname=b"Andrew Nesbitt <andrewnez@gmail.com>",
        email=b"andrewnez@gmail.com",
    ),
    synthetic=False,
    date=TimestampWithTimezone.from_datetime(
        datetime.datetime(2017, 2, 20, 16, 14, 16, tzinfo=_utc_plus_2)
    ),
    target_type=ObjectType.DIRECTORY,
    target=DIRECTORY2.id,
)

RELEASES = [RELEASE]

SNAPSHOTS = [
    # https://github.com/SoftwareHeritage/swh-storage
    Snapshot(
        branches={
            b"refs/heads/add-revision-origin-cache": SnapshotBranch(
                target=b'L[\xce\x1c\x88\x8eF\t\xf1"\x19\x1e\xfb\xc0s\xe7/\xe9l\x1e',
                target_type=SnapshotTargetType.REVISION,
            ),
            b"refs/head/master": SnapshotBranch(
                target=b"8K\x12\x00d\x03\xcc\xe4]bS\xe3\x8f{\xd7}\xac\xefrm",
                target_type=SnapshotTargetType.REVISION,
            ),
            b"HEAD": SnapshotBranch(
                target=b"refs/head/master", target_type=SnapshotTargetType.ALIAS
            ),
            b"refs/tags/v0.0.103": SnapshotBranch(
                target=b'\xb6"Im{\xfdLb\xb0\x94N\xea\x96m\x13x\x88+\x0f\xdd',
                target_type=SnapshotTargetType.RELEASE,
            ),
        },
    ),
    # rsync://ftp.gnu.org/gnu/3dldf
    Snapshot(
        branches={
            b"3DLDF-1.1.4.tar.gz": SnapshotBranch(
                target=b'dJ\xfb\x1c\x91\xf4\x82B%]6\xa2\x90|\xd3\xfc"G\x99\x11',
                target_type=SnapshotTargetType.REVISION,
            ),
            b"3DLDF-2.0.2.tar.gz": SnapshotBranch(
                target=b"\xb6\x0e\xe7\x9e9\xac\xaa\x19\x9e=\xd1\xc5\x00\\\xc6\xfc\xe0\xa6\xb4V",  # noqa
                target_type=SnapshotTargetType.REVISION,
            ),
            b"3DLDF-2.0.3-examples.tar.gz": SnapshotBranch(
                target=b"!H\x19\xc0\xee\x82-\x12F1\xbd\x97\xfe\xadZ\x80\x80\xc1\x83\xff",  # noqa
                target_type=SnapshotTargetType.REVISION,
            ),
            b"3DLDF-2.0.3.tar.gz": SnapshotBranch(
                target=b"\x8e\xa9\x8e/\xea}\x9feF\xf4\x9f\xfd\xee\xcc\x1a\xb4`\x8c\x8by",  # noqa
                target_type=SnapshotTargetType.REVISION,
            ),
            b"3DLDF-2.0.tar.gz": SnapshotBranch(
                target=b"F6*\xff(?\x19a\xef\xb6\xc2\x1fv$S\xe3G\xd3\xd1m",
                target_type=SnapshotTargetType.REVISION,
            ),
        },
    ),
    # https://forge.softwareheritage.org/source/jesuisgpl/",
    Snapshot(
        branches={
            b"master": SnapshotBranch(
                target=b"\xe7n\xa4\x9c\x9f\xfb\xb7\xf76\x11\x08{\xa6\xe9\x99\xb1\x9e]q\xeb",  # noqa
                target_type=SnapshotTargetType.REVISION,
            )
        },
    ),
    # https://old-pypi.example.org/project/limnoria/
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=b"releases/2018.09.09", target_type=SnapshotTargetType.ALIAS
            ),
            b"releases/2018.09.01": SnapshotBranch(
                target=b"<\xee1(\xe8\x8d_\xc1\xc9\xa6rT\xf1\x1d\xbb\xdfF\xfdw\xcf",
                target_type=SnapshotTargetType.REVISION,
            ),
            b"releases/2018.09.09": SnapshotBranch(
                target=b"\x83\xb9\xb6\xc7\x05\xb1%\xd0\xfem\xd8kA\x10\x9d\xc5\xfa2\xf8t",  # noqa
                target_type=SnapshotTargetType.REVISION,
            ),
        },
    ),
    # https://pypi.org/project/limnoria/
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=b"releases/2018.09.09", target_type=SnapshotTargetType.ALIAS
            ),
            b"releases/2018.09.01": SnapshotBranch(
                target=b"<\xee1(\xe8\x8d_\xc1\xc9\xa6rT\xf1\x1d\xbb\xdfF\xfdw\xcf",
                target_type=SnapshotTargetType.RELEASE,
            ),
            b"releases/2018.09.09": SnapshotBranch(
                target=b"\x83\xb9\xb6\xc7\x05\xb1%\xd0\xfem\xd8kA\x10\x9d\xc5\xfa2\xf8t",  # noqa
                target_type=SnapshotTargetType.RELEASE,
            ),
        },
    ),
    # http://0-512-md.googlecode.com/svn/
    Snapshot(
        branches={
            b"master": SnapshotBranch(
                target=b"\xe4?r\xe1,\x88\xab\xec\xe7\x9a\x87\xb8\xc9\xad#.\x1bw=\x18",
                target_type=SnapshotTargetType.REVISION,
            )
        },
    ),
    # https://github.com/librariesio/yarn-parser
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=REVISION.id,
                target_type=SnapshotTargetType.REVISION,
            )
        },
    ),
    # https://github.com/librariesio/yarn-parser.git
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=REVISION.id,
                target_type=SnapshotTargetType.REVISION,
            )
        },
    ),
    # https://npm.example.org/yarn-parser
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=RELEASE.id,
                target_type=SnapshotTargetType.RELEASE,
            )
        },
    ),
]

assert len(SNAPSHOTS) == len(ORIGIN_VISITS)


YARN_PARSER_METADATA = {
    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
    "url": "https://github.com/librariesio/yarn-parser#readme",
    "codeRepository": "git+git+https://github.com/librariesio/yarn-parser.git",
    "author": [{"type": "Person", "name": "Andrew Nesbitt"}],
    "license": "https://spdx.org/licenses/AGPL-3.0",
    "version": "1.0.0",
    "description": "Tiny web service for parsing yarn.lock files",
    "issueTracker": "https://github.com/librariesio/yarn-parser/issues",
    "name": "yarn-parser",
    "type": "SoftwareSourceCode",
}


json_dict_keys = strategies.one_of(
    strategies.characters(),
    strategies.just("type"),
    strategies.just("url"),
    strategies.just("name"),
    strategies.just("email"),
    strategies.just("@id"),
    strategies.just("@context"),
    strategies.just("repository"),
    strategies.just("license"),
    strategies.just("repositories"),
    strategies.just("licenses"),
)
"""Hypothesis strategy that generates strings, with an emphasis on those
that are often used as dictionary keys in metadata files."""


generic_json_document = strategies.recursive(
    strategies.none()
    | strategies.booleans()
    | strategies.floats()
    | strategies.characters(),
    lambda children: (
        strategies.lists(children, min_size=1)
        | strategies.dictionaries(json_dict_keys, children, min_size=1)
    ),
)
"""Hypothesis strategy that generates possible values for values of JSON
metadata files."""


def json_document_strategy(keys=None):
    """Generates an hypothesis strategy that generates metadata files
    for a JSON-based format that uses the given keys."""
    if keys is None:
        keys = strategies.characters()
    else:
        keys = strategies.one_of(map(strategies.just, keys))

    return strategies.dictionaries(keys, generic_json_document, min_size=1)


def _tree_to_xml(root, xmlns, data):
    def encode(s):
        "Skips unpaired surrogates generated by json_document_strategy"
        return s.encode("utf8", "replace")

    def to_xml(data, indent=b" "):
        if data is None:
            return b""
        elif isinstance(data, (bool, str, int, float)):
            return indent + encode(str(data))
        elif isinstance(data, list):
            return b"\n".join(to_xml(v, indent=indent) for v in data)
        elif isinstance(data, dict):
            lines = []
            for key, value in data.items():
                lines.append(indent + encode("<{}>".format(key)))
                lines.append(to_xml(value, indent=indent + b" "))
                lines.append(indent + encode("</{}>".format(key)))
            return b"\n".join(lines)
        else:
            raise TypeError(data)

    return b"\n".join(
        [
            '<{} xmlns="{}">'.format(root, xmlns).encode(),
            to_xml(data),
            "</{}>".format(root).encode(),
        ]
    )


class TreeToXmlTest(unittest.TestCase):
    def test_leaves(self):
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", None),
            b'<root xmlns="http://example.com">\n\n</root>',
        )
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", True),
            b'<root xmlns="http://example.com">\n True\n</root>',
        )
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", "abc"),
            b'<root xmlns="http://example.com">\n abc\n</root>',
        )
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", 42),
            b'<root xmlns="http://example.com">\n 42\n</root>',
        )
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", 3.14),
            b'<root xmlns="http://example.com">\n 3.14\n</root>',
        )

    def test_dict(self):
        self.assertIn(
            _tree_to_xml("root", "http://example.com", {"foo": "bar", "baz": "qux"}),
            [
                b'<root xmlns="http://example.com">\n'
                b" <foo>\n  bar\n </foo>\n"
                b" <baz>\n  qux\n </baz>\n"
                b"</root>",
                b'<root xmlns="http://example.com">\n'
                b" <baz>\n  qux\n </baz>\n"
                b" <foo>\n  bar\n </foo>\n"
                b"</root>",
            ],
        )

    def test_list(self):
        self.assertEqual(
            _tree_to_xml(
                "root",
                "http://example.com",
                [
                    {"foo": "bar"},
                    {"foo": "baz"},
                ],
            ),
            b'<root xmlns="http://example.com">\n'
            b" <foo>\n  bar\n </foo>\n"
            b" <foo>\n  baz\n </foo>\n"
            b"</root>",
        )


def xml_document_strategy(keys, root, xmlns):
    """Generates an hypothesis strategy that generates metadata files
    for an XML format that uses the given keys."""

    return strategies.builds(
        functools.partial(_tree_to_xml, root, xmlns), json_document_strategy(keys)
    )


def filter_dict(d, keys):
    "return a copy of the dict with keys deleted"
    if not isinstance(keys, (list, tuple)):
        keys = (keys,)
    return dict((k, v) for (k, v) in d.items() if k not in keys)


def fill_obj_storage(obj_storage):
    """Add some content in an object storage."""
    for obj_id, content in OBJ_STORAGE_DATA.items():
        obj_storage.add(content, obj_id)


def fill_storage(storage):
    """Fill in storage with consistent test dataset."""
    storage.content_add([Content.from_data(data) for data in OBJ_STORAGE_DATA.values()])
    storage.directory_add([DIRECTORY, DIRECTORY2])
    storage.revision_add(REVISIONS)
    storage.release_add(RELEASES)
    storage.snapshot_add(SNAPSHOTS)

    storage.origin_add(ORIGINS)
    for visit, snapshot in zip(ORIGIN_VISITS, SNAPSHOTS):
        assert snapshot.id is not None

        visit = storage.origin_visit_add(
            [OriginVisit(origin=visit["origin"], date=now(), type=visit["type"])]
        )[0]
        visit_status = OriginVisitStatus(
            origin=visit.origin,
            visit=visit.visit,
            date=now(),
            status="full",
            snapshot=snapshot.id,
        )
        storage.origin_visit_status_add([visit_status])


class CommonContentIndexerTest(metaclass=abc.ABCMeta):
    def get_indexer_results(self, ids):
        """Override this for indexers that don't have a mock storage."""
        return self.indexer.idx_storage.state

    def assert_results_ok(self, sha1s, expected_results=None):
        sha1s = [hash_to_bytes(sha1) for sha1 in sha1s]
        actual_results = list(self.get_indexer_results(sha1s))

        if expected_results is None:
            expected_results = self.expected_results

        # expected results may contain slightly duplicated results
        assert 0 < len(actual_results) <= len(expected_results)
        for result in actual_results:
            assert result in expected_results

    def test_index(self):
        """Known sha1 have their data indexed"""
        sha1s = [self.id0, self.id1, self.id2]

        # when
        self.indexer.run(sha1s)

        self.assert_results_ok(sha1s)

        # 2nd pass
        self.indexer.run(sha1s)

        self.assert_results_ok(sha1s)

    def test_index_one_unknown_sha1(self):
        """Unknown sha1s are not indexed"""
        sha1s = [
            self.id1,
            "799a5ef812c53907562fe379d4b3851e69c7cb15",  # unknown
            "800a5ef812c53907562fe379d4b3851e69c7cb15",  # unknown
        ]  # unknown

        # when
        self.indexer.run(sha1s)

        # then
        expected_results = [res for res in self.expected_results if res.id in sha1s]

        self.assert_results_ok(sha1s, expected_results)


class CommonContentIndexerPartitionTest:
    """Allows to factorize tests on range indexer."""

    def setUp(self):
        self.contents = sorted(OBJ_STORAGE_DATA)

    def assert_results_ok(self, partition_id, nb_partitions, actual_results):
        expected_ids = [
            c.sha1
            for c in stream_results(
                self.indexer.storage.content_get_partition,
                partition_id=partition_id,
                nb_partitions=nb_partitions,
            )
        ]

        actual_results = list(actual_results)
        for indexed_data in actual_results:
            _id = indexed_data.id
            assert _id in expected_ids

            _tool_id = indexed_data.indexer_configuration_id
            assert _tool_id == self.indexer.tool["id"]

    def test__index_contents(self):
        """Indexing contents without existing data results in indexed data"""
        partition_id = 0
        nb_partitions = 4

        actual_results = list(
            self.indexer._index_contents(partition_id, nb_partitions, indexed={})
        )

        self.assert_results_ok(partition_id, nb_partitions, actual_results)

    def test__index_contents_with_indexed_data(self):
        """Indexing contents with existing data results in less indexed data"""
        partition_id = 3
        nb_partitions = 4

        # first pass
        actual_results = list(
            self.indexer._index_contents(partition_id, nb_partitions, indexed={}),
        )

        self.assert_results_ok(partition_id, nb_partitions, actual_results)

        indexed_ids = {res.id for res in actual_results}

        actual_results = list(
            self.indexer._index_contents(
                partition_id, nb_partitions, indexed=indexed_ids
            )
        )

        # already indexed, so nothing new
        assert actual_results == []

    def test_generate_content_get(self):
        """Optimal indexing should result in indexed data"""
        partition_id = 0
        nb_partitions = 1

        actual_results = self.indexer.run(
            partition_id, nb_partitions, skip_existing=False
        )

        assert actual_results["status"] == "eventful", actual_results

    def test_generate_content_get_no_result(self):
        """No result indexed returns False"""
        actual_results = self.indexer.run(1, 2**512, incremental=False)

        assert actual_results == {"status": "uneventful"}


def mock_compute_license(path):
    """path is the content identifier"""
    if isinstance(id, bytes):
        path = path.decode("utf-8")
    # path is something like /tmp/tmpXXX/<sha1> so we keep only the sha1 part
    id_ = path.split("/")[-1]
    return {"licenses": SHA1_TO_LICENSES.get(hash_to_bytes(id_), [])}



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
from typing import Dict, Type

import click

from . import (
    cff,
    codemeta,
    composer,
    dart,
    gitea,
    github,
    maven,
    npm,
    nuget,
    python,
    ruby,
)
from .base import BaseExtrinsicMapping, BaseIntrinsicMapping, BaseMapping

INTRINSIC_MAPPINGS: Dict[str, Type[BaseIntrinsicMapping]] = {
    "CffMapping": cff.CffMapping,
    "CodemetaMapping": codemeta.CodemetaMapping,
    "GemspecMapping": ruby.GemspecMapping,
    "MavenMapping": maven.MavenMapping,
    "NpmMapping": npm.NpmMapping,
    "PubMapping": dart.PubspecMapping,
    "PythonPkginfoMapping": python.PythonPkginfoMapping,
    "ComposerMapping": composer.ComposerMapping,
    "NuGetMapping": nuget.NuGetMapping,
}

EXTRINSIC_MAPPINGS: Dict[str, Type[BaseExtrinsicMapping]] = {
    "GiteaMapping": gitea.GiteaMapping,
    "GitHubMapping": github.GitHubMapping,
    "JsonSwordCodemetaMapping": codemeta.JsonSwordCodemetaMapping,
    "SwordCodemetaMapping": codemeta.SwordCodemetaMapping,
}


MAPPINGS: Dict[str, Type[BaseMapping]] = {**INTRINSIC_MAPPINGS, **EXTRINSIC_MAPPINGS}


def list_terms():
    """Returns a dictionary with all supported CodeMeta terms as keys,
    and the mappings that support each of them as values."""
    d = collections.defaultdict(set)
    for mapping in MAPPINGS.values():
        for term in mapping.supported_terms():
            d[term].add(mapping)
    return d


@click.command()
@click.argument("mapping_name")
@click.argument("file_name")
def main(mapping_name: str, file_name: str):
    from pprint import pprint

    with open(file_name, "rb") as fd:
        file_content = fd.read()
    res = MAPPINGS[mapping_name]().translate(file_content)
    pprint(res)


if __name__ == "__main__":
    main()



# Copyright (C) 2017-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from copy import deepcopy
import datetime
import hashlib
import logging
import re
import time
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)
import urllib.parse
from urllib.parse import urlparse

import pkg_resources
import sentry_sdk

from swh.core.config import merge_configs
from swh.core.utils import grouper
from swh.indexer.codemeta import merge_documents
from swh.indexer.indexer import (
    BaseIndexer,
    ContentIndexer,
    DirectoryIndexer,
    ObjectsDict,
    OriginIndexer,
)
from swh.indexer.metadata_detector import detect_metadata
from swh.indexer.metadata_dictionary import EXTRINSIC_MAPPINGS, INTRINSIC_MAPPINGS
from swh.indexer.metadata_dictionary.base import DirectoryLsEntry
from swh.indexer.origin_head import get_head_swhid
from swh.indexer.storage import INDEXER_CFG_KEY, Sha1
from swh.indexer.storage.model import (
    ContentMetadataRow,
    DirectoryIntrinsicMetadataRow,
    OriginExtrinsicMetadataRow,
    OriginIntrinsicMetadataRow,
)
from swh.model import hashutil
from swh.model.model import (
    Directory,
    MetadataAuthorityType,
    Origin,
    RawExtrinsicMetadata,
    ReleaseTargetType,
    Sha1Git,
)
from swh.model.swhids import CoreSWHID, ExtendedObjectType, ObjectType

REVISION_GET_BATCH_SIZE = 10
RELEASE_GET_BATCH_SIZE = 10
ORIGIN_GET_BATCH_SIZE = 10


T1 = TypeVar("T1")
T2 = TypeVar("T2")

logger = logging.getLogger(__name__)


def call_with_batches(
    f: Callable[[List[T1]], Iterable[T2]],
    args: List[T1],
    batch_size: int,
) -> Iterator[T2]:
    """Calls a function with batches of args, and concatenates the results."""
    groups = grouper(args, batch_size)
    for group in groups:
        yield from f(list(group))


class ExtrinsicMetadataIndexer(
    BaseIndexer[Sha1Git, RawExtrinsicMetadata, OriginExtrinsicMetadataRow]
):
    def process_journal_objects(self, objects: ObjectsDict) -> Dict:
        summary: Dict[str, Any] = {"status": "uneventful"}
        try:
            results = {}
            for item in objects.get("raw_extrinsic_metadata", []):
                remd = RawExtrinsicMetadata.from_dict(item)
                sentry_sdk.set_tag("swh-indexer-remd-swhid", str(remd.swhid()))
                for result in self.index(remd.id, data=remd):
                    results[result.id] = result
        except Exception:
            if not self.catch_exceptions:
                raise
            summary["status"] = "failed"
            return summary

        self.results = list(results.values())
        summary_persist = self.persist_index_computations(self.results)
        if summary_persist:
            for value in summary_persist.values():
                if value > 0:
                    summary["status"] = "eventful"
            summary.update(summary_persist)
        return summary

    def index(
        self,
        id: Sha1Git,
        data: Optional[RawExtrinsicMetadata],
        **kwargs,
    ) -> List[OriginExtrinsicMetadataRow]:
        if data is None:
            raise NotImplementedError(
                "ExtrinsicMetadataIndexer.index() without RawExtrinsicMetadata data"
            )
        if data.target.object_type == ExtendedObjectType.ORIGIN:
            origin_sha1 = data.target.object_id
        elif data.origin is not None:
            if (
                data.fetcher.name == "swh-deposit"
                and data.discovery_date
                < datetime.datetime(
                    2024, 5, 16, 15, 20, 0, tzinfo=datetime.timezone.utc
                )
            ):
                # Workaround for deposits while swh.model.swhids did not unescape origin
                # URLs while parsing <swh:object swhid="..." />, which was fixed in
                # https://gitlab.softwareheritage.org/swh/devel/swh-model/-/merge_requests/348
                # itself deployed shortly after 2024-05-16T15:18:07 by
                # https://gitlab.softwareheritage.org/swh/infra/swh-apps/-/commit/70bd86aafcbc1787183e5d2cd52c392ae012e65e
                if "%" in data.origin:
                    assert re.match(
                        "^https://cran.r-project.org/package%3D[a-zA-Z]+$", data.origin
                    ), data
                origin_url = urllib.parse.unquote_to_bytes(data.origin)
            else:
                origin_url = data.origin.encode()

            # HACK: As swh-search does not (yet?) support searching on directories
            # and traversing back to origins, we index metadata on non-origins with
            # an origin context as if they were on the origin itself.
            origin_sha1 = hashlib.sha1(origin_url).digest()
        else:
            # other types are not supported yet
            return []

        if data.authority.type == MetadataAuthorityType.REGISTRY:
            # metadata provided by a third-party; don't trust it
            # (technically this could be handled below, but we check it here
            # to return early; sparing a translation and origin lookup)
            # TODO: add ways to define trusted authorities
            return []

        metadata_items = []
        mappings: List[str] = []
        for mapping_cls in EXTRINSIC_MAPPINGS.values():
            if data.format in mapping_cls.extrinsic_metadata_formats():
                mapping = mapping_cls()
                metadata_item = mapping.translate(data.metadata)
                if metadata_item is not None:
                    metadata_items.append(metadata_item)
                    mappings.append(mapping.name)

        if not metadata_items:
            # Don't have any mapping to parse it, ignore
            return []

        # TODO: batch requests to origin_get_by_sha1()
        num_retries = 6
        sleep_delay = 10
        for _ in range(num_retries):
            origins = self.storage.origin_get_by_sha1([origin_sha1])
            try:
                (origin,) = origins
                if origin is not None:
                    break
            except ValueError:
                pass
            # The origin does not exist. This may be due to some replication lag
            # between the loader's DB/journal and the DB we are consuming from.
            # Wait a bit and try again
            logger.debug(
                "Origin %s not found, sleeping for %ss.", data.target, sleep_delay
            )
            time.sleep(sleep_delay)
        else:
            # Does not exist, or replication lag > 60s.
            raise ValueError(
                f"Unknown origin swh:1:ori:{origin_sha1.hex()} for metadata target: "
                f"{data.target}. Is the swh-storage database replication lag "
                f"over {num_retries*sleep_delay}s?"
            ) from None

        if urlparse(data.authority.url).netloc != urlparse(origin["url"]).netloc:
            # metadata provided by a third-party; don't trust it
            # TODO: add ways to define trusted authorities
            return []

        metadata = merge_documents(metadata_items)

        return [
            OriginExtrinsicMetadataRow(
                id=origin["url"],
                indexer_configuration_id=self.tool["id"],
                from_remd_id=data.id,
                mappings=mappings,
                metadata=metadata,
            )
        ]

    def persist_index_computations(
        self, results: List[OriginExtrinsicMetadataRow]
    ) -> Dict[str, int]:
        """Persist the results in storage."""
        return self.idx_storage.origin_extrinsic_metadata_add(results)


class ContentMetadataIndexer(ContentIndexer[ContentMetadataRow]):
    """Content-level indexer

    This indexer is in charge of:

    - filtering out content already indexed in content_metadata
    - reading content from objstorage with the content's id sha1
    - computing metadata by given context
    - using the metadata_dictionary as the 'swh-metadata-translator' tool
    - store result in content_metadata table

    """

    def filter(self, ids):
        """Filter out known sha1s and return only missing ones."""
        yield from self.idx_storage.content_metadata_missing(
            (
                {
                    "id": sha1,
                    "indexer_configuration_id": self.tool["id"],
                }
                for sha1 in ids
            )
        )

    def index(
        self,
        id: Sha1,
        data: Optional[bytes] = None,
        log_suffix="unknown directory",
        **kwargs,
    ) -> List[ContentMetadataRow]:
        """Index sha1s' content and store result.

        Args:
            id: content's identifier
            data: raw content in bytes

        Returns:
            dict: dictionary representing a content_metadata. If the
            translation wasn't successful the metadata keys will
            be returned as None

        """
        assert isinstance(id, bytes)
        assert data is not None
        metadata = None
        try:
            mapping_name = self.tool["tool_configuration"]["context"]
            log_suffix += ", content_id=%s" % hashutil.hash_to_hex(id)
            metadata = INTRINSIC_MAPPINGS[mapping_name](log_suffix).translate(data)
        except Exception:
            self.log.exception(
                "Problem during metadata translation "
                "for content %s" % hashutil.hash_to_hex(id)
            )
            sentry_sdk.capture_exception()
        if metadata is None:
            return []
        return [
            ContentMetadataRow(
                id=id,
                indexer_configuration_id=self.tool["id"],
                metadata=metadata,
            )
        ]

    def persist_index_computations(
        self, results: List[ContentMetadataRow]
    ) -> Dict[str, int]:
        """Persist the results in storage."""
        return self.idx_storage.content_metadata_add(results)


DEFAULT_CONFIG: Dict[str, Any] = {
    "tools": {
        "name": "swh.indexer.metadata",
        "version": pkg_resources.get_distribution("swh.indexer").version,
        "configuration": {},
    },
}


class DirectoryMetadataIndexer(DirectoryIndexer[DirectoryIntrinsicMetadataRow]):
    """Directory-level indexer

    This indexer is in charge of:

    - filtering directories already indexed in directory_intrinsic_metadata table
      with defined computation tool
    - retrieve all entry_files in directory
    - use metadata_detector for file_names containing metadata
    - compute metadata translation if necessary and possible (depends on tool)
    - send sha1s to content indexing if possible
    - store the results for directory

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = merge_configs(DEFAULT_CONFIG, self.config)

    def filter(self, sha1_gits):
        """Filter out known sha1s and return only missing ones."""
        yield from self.idx_storage.directory_intrinsic_metadata_missing(
            (
                {
                    "id": sha1_git,
                    "indexer_configuration_id": self.tool["id"],
                }
                for sha1_git in sha1_gits
            )
        )

    def index(
        self, id: Sha1Git, data: Optional[Directory] = None, **kwargs
    ) -> List[DirectoryIntrinsicMetadataRow]:
        """Index directory by processing it and organizing result.

        use metadata_detector to iterate on filenames, passes them to the content
        indexers, then merges (if more than one)

        Args:
          id: sha1_git of the directory
          data: should always be None

        Returns:
            dict: dictionary representing a directory_intrinsic_metadata, with
            keys:

            - id: directory's identifier (sha1_git)
            - indexer_configuration_id (bytes): tool used
            - metadata: dict of retrieved metadata

        """
        dir_: List[DirectoryLsEntry]
        assert data is None, "Unexpected directory object"
        dir_ = cast(
            List[DirectoryLsEntry],
            list(self.storage.directory_ls(id, recursive=False)),
        )

        try:
            if [entry["type"] for entry in dir_] == ["dir"]:
                # If the root is just a single directory, recurse into it
                # eg. PyPI packages, GNU tarballs
                subdir = dir_[0]["target"]
                dir_ = cast(
                    List[DirectoryLsEntry],
                    list(self.storage.directory_ls(subdir, recursive=False)),
                )
            files = [entry for entry in dir_ if entry["type"] == "file"]
            (mappings, metadata) = self.translate_directory_intrinsic_metadata(
                files,
                log_suffix="directory=%s" % hashutil.hash_to_hex(id),
            )
        except Exception as e:
            self.log.exception("Problem when indexing dir: %r", e)
            sentry_sdk.capture_exception()
            return []
        return [
            DirectoryIntrinsicMetadataRow(
                id=id,
                indexer_configuration_id=self.tool["id"],
                mappings=mappings,
                metadata=metadata,
            )
        ]

    def persist_index_computations(
        self, results: List[DirectoryIntrinsicMetadataRow]
    ) -> Dict[str, int]:
        """Persist the results in storage."""
        # TODO: add functions in storage to keep data in
        # directory_intrinsic_metadata
        return self.idx_storage.directory_intrinsic_metadata_add(results)

    def translate_directory_intrinsic_metadata(
        self, files: List[DirectoryLsEntry], log_suffix: str
    ) -> Tuple[List[Any], Any]:
        """
        Determine plan of action to translate metadata in the given root directory

        Args:
            files: list of file entries, as returned by
              :meth:`swh.storage.interface.StorageInterface.directory_ls`

        Returns:
            (List[str], dict): list of mappings used and dict with
            translated metadata according to the CodeMeta vocabulary

        """
        metadata = []
        # TODO: iterate on each context, on each file
        # -> get raw_contents
        # -> translate each content
        config = {
            k: self.config[k]
            for k in [INDEXER_CFG_KEY, "objstorage", "storage", "tools"]
        }
        all_detected_files = detect_metadata(files)
        used_mappings = [
            INTRINSIC_MAPPINGS[context].name for context in all_detected_files
        ]
        for mapping_name, detected_files in all_detected_files.items():
            cfg = deepcopy(config)
            cfg["tools"]["configuration"]["context"] = mapping_name
            c_metadata_indexer = ContentMetadataIndexer(config=cfg)
            # sha1s that are in content_metadata table
            sha1s_in_storage = []
            metadata_generator = self.idx_storage.content_metadata_get(detected_files)
            for c in metadata_generator:
                # extracting metadata
                sha1 = c.id
                sha1s_in_storage.append(sha1)
                local_metadata = c.metadata
                # local metadata is aggregated
                if local_metadata:
                    metadata.append(local_metadata)

            sha1s_filtered = [
                item for item in detected_files if item not in sha1s_in_storage
            ]

            if sha1s_filtered:
                # content indexing
                try:
                    c_metadata_indexer.run(
                        sha1s_filtered,
                        log_suffix=log_suffix,
                    )
                    # on the fly possibility:
                    for result in c_metadata_indexer.results:
                        local_metadata = result.metadata
                        metadata.append(local_metadata)

                except Exception:
                    self.log.exception("Exception while indexing metadata on contents")
                    sentry_sdk.capture_exception()

        metadata = merge_documents(metadata)
        return (used_mappings, metadata)


class OriginMetadataIndexer(
    OriginIndexer[Tuple[OriginIntrinsicMetadataRow, DirectoryIntrinsicMetadataRow]]
):
    USE_TOOLS = False

    def __init__(self, config=None, **kwargs) -> None:
        super().__init__(config=config, **kwargs)
        self.directory_metadata_indexer = DirectoryMetadataIndexer(config=config)

    def index_list(
        self,
        origins: List[Origin],
        *,
        check_origin_known: bool = True,
        **kwargs,
    ) -> List[Tuple[OriginIntrinsicMetadataRow, DirectoryIntrinsicMetadataRow]]:
        head_rev_ids = []
        head_rel_ids = []
        origin_heads: Dict[Origin, CoreSWHID] = {}

        # Filter out origins not in the storage
        if check_origin_known:
            known_origins = list(
                call_with_batches(
                    self.storage.origin_get,
                    [origin.url for origin in origins],
                    ORIGIN_GET_BATCH_SIZE,
                )
            )
        else:
            known_origins = list(origins)

        for origin in known_origins:
            if origin is None:
                continue
            head_swhid = get_head_swhid(self.storage, origin.url)
            if head_swhid:
                origin_heads[origin] = head_swhid
                if head_swhid.object_type == ObjectType.REVISION:
                    head_rev_ids.append(head_swhid.object_id)
                elif head_swhid.object_type == ObjectType.RELEASE:
                    head_rel_ids.append(head_swhid.object_id)
                else:
                    assert False, head_swhid

        head_revs = dict(
            zip(
                head_rev_ids,
                call_with_batches(
                    self.storage.revision_get, head_rev_ids, REVISION_GET_BATCH_SIZE
                ),
            )
        )
        head_rels = dict(
            zip(
                head_rel_ids,
                call_with_batches(
                    self.storage.release_get, head_rel_ids, RELEASE_GET_BATCH_SIZE
                ),
            )
        )

        results = []
        for origin, head_swhid in origin_heads.items():
            sentry_sdk.set_tag("swh-indexer-origin-url", origin.url)
            sentry_sdk.set_tag("swh-indexer-origin-head-swhid", str(head_swhid))
            if head_swhid.object_type == ObjectType.REVISION:
                rev = head_revs[head_swhid.object_id]
                if not rev:
                    self.log.warning(
                        "Missing head object %s of origin %r", head_swhid, origin.url
                    )
                    continue
                directory_id = rev.directory
            elif head_swhid.object_type == ObjectType.RELEASE:
                rel = head_rels[head_swhid.object_id]
                if not rel:
                    self.log.warning(
                        "Missing head object %s of origin %r", head_swhid, origin.url
                    )
                    continue
                if rel.target_type != ReleaseTargetType.DIRECTORY:
                    # TODO
                    self.log.warning(
                        "Head release %s of %r has unexpected target type %s",
                        head_swhid,
                        origin.url,
                        rel.target_type,
                    )
                    continue
                assert rel.target, rel
                directory_id = rel.target
            else:
                assert False, head_swhid

            for dir_metadata in self.directory_metadata_indexer.index(directory_id):
                # There is at most one dir_metadata
                orig_metadata = OriginIntrinsicMetadataRow(
                    from_directory=dir_metadata.id,
                    id=origin.url,
                    metadata=dir_metadata.metadata,
                    mappings=dir_metadata.mappings,
                    indexer_configuration_id=dir_metadata.indexer_configuration_id,
                )
                results.append((orig_metadata, dir_metadata))

        return results

    def persist_index_computations(
        self,
        results: List[Tuple[OriginIntrinsicMetadataRow, DirectoryIntrinsicMetadataRow]],
    ) -> Dict[str, int]:
        # Deduplicate directories
        dir_metadata: Dict[bytes, DirectoryIntrinsicMetadataRow] = {}
        orig_metadata: Dict[str, OriginIntrinsicMetadataRow] = {}
        summary: Dict = {}
        for orig_item, dir_item in results:
            assert dir_item.metadata == orig_item.metadata
            if dir_item.metadata and not (dir_item.metadata.keys() <= {"@context"}):
                # Only store non-empty metadata sets
                if dir_item.id not in dir_metadata:
                    dir_metadata[dir_item.id] = dir_item
                if orig_item.id not in orig_metadata:
                    orig_metadata[orig_item.id] = orig_item

        if dir_metadata:
            summary_dir = self.idx_storage.directory_intrinsic_metadata_add(
                list(dir_metadata.values())
            )
            summary.update(summary_dir)
        if orig_metadata:
            summary_ori = self.idx_storage.origin_intrinsic_metadata_add(
                list(orig_metadata.values())
            )
            summary.update(summary_ori)

        return summary



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.indexer.codemeta import CROSSWALK_TABLE, merge_documents


def test_crosstable():
    assert {k: str(v) for (k, v) in CROSSWALK_TABLE["NodeJS"].items()} == {
        "repository": "http://schema.org/codeRepository",
        "os": "http://schema.org/operatingSystem",
        "cpu": "http://schema.org/processorRequirements",
        "engines": "http://schema.org/runtimePlatform",
        "author": "http://schema.org/author",
        "author.email": "http://schema.org/email",
        "author.name": "http://schema.org/name",
        "contributors": "http://schema.org/contributor",
        "keywords": "http://schema.org/keywords",
        "license": "http://schema.org/license",
        "version": "http://schema.org/version",
        "description": "http://schema.org/description",
        "name": "http://schema.org/name",
        "bugs": "https://codemeta.github.io/terms/issueTracker",
        "homepage": "http://schema.org/url",
    }


def test_merge_documents():
    """
    Test the creation of a coherent minimal metadata set
    """
    # given
    metadata_list = [
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "name": "test_1",
            "version": "0.0.2",
            "description": "Simple package.json test for indexer",
            "codeRepository": "git+https://github.com/moranegg/metadata_test",
        },
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "name": "test_0_1",
            "version": "0.0.2",
            "description": "Simple package.json test for indexer",
            "codeRepository": "git+https://github.com/moranegg/metadata_test",
        },
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "name": "test_metadata",
            "version": "0.0.2",
            "author": {
                "type": "Person",
                "name": "moranegg",
            },
        },
    ]

    # when
    results = merge_documents(metadata_list)

    # then
    expected_results = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "version": "0.0.2",
        "description": "Simple package.json test for indexer",
        "name": ["test_1", "test_0_1", "test_metadata"],
        "author": [{"type": "Person", "name": "moranegg"}],
        "codeRepository": "git+https://github.com/moranegg/metadata_test",
    }
    assert results == expected_results


def test_merge_documents_ids():
    # given
    metadata_list = [
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "id": "http://example.org/test1",
            "name": "test_1",
        },
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "id": "http://example.org/test2",
            "name": "test_2",
        },
    ]

    # when
    results = merge_documents(metadata_list)

    # then
    expected_results = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "id": "http://example.org/test1",
        "schema:sameAs": "http://example.org/test2",
        "name": ["test_1", "test_2"],
    }
    assert results == expected_results


def test_merge_documents_duplicate_ids():
    # given
    metadata_list = [
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "id": "http://example.org/test1",
            "name": "test_1",
        },
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "id": "http://example.org/test1",
            "name": "test_1b",
        },
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "id": "http://example.org/test2",
            "name": "test_2",
        },
    ]

    # when
    results = merge_documents(metadata_list)

    # then
    expected_results = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "id": "http://example.org/test1",
        "schema:sameAs": "http://example.org/test2",
        "name": ["test_1", "test_1b", "test_2"],
    }
    assert results == expected_results


def test_merge_documents_lists():
    """Tests merging two @list elements."""
    # given
    metadata_list = [
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {
                "@list": [
                    {"name": "test_1"},
                ]
            },
        },
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {
                "@list": [
                    {"name": "test_2"},
                ]
            },
        },
    ]

    # when
    results = merge_documents(metadata_list)

    # then
    expected_results = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "author": [
            {"name": "test_1"},
            {"name": "test_2"},
        ],
    }
    assert results == expected_results


def test_merge_documents_lists_duplicates():
    """Tests merging two @list elements with a duplicate subelement."""
    # given
    metadata_list = [
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {
                "@list": [
                    {"name": "test_1"},
                ]
            },
        },
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {
                "@list": [
                    {"name": "test_2"},
                    {"name": "test_1"},
                ]
            },
        },
    ]

    # when
    results = merge_documents(metadata_list)

    # then
    expected_results = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "author": [
            {"name": "test_1"},
            {"name": "test_2"},
        ],
    }
    assert results == expected_results


def test_merge_documents_list_left():
    """Tests merging a singleton with an @list."""
    # given
    metadata_list = [
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {"name": "test_1"},
        },
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {
                "@list": [
                    {"name": "test_2"},
                ]
            },
        },
    ]

    # when
    results = merge_documents(metadata_list)

    # then
    expected_results = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "author": [
            {"name": "test_1"},
            {"name": "test_2"},
        ],
    }
    assert results == expected_results


def test_merge_documents_list_right():
    """Tests merging an @list with a singleton."""
    # given
    metadata_list = [
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {
                "@list": [
                    {"name": "test_1"},
                ]
            },
        },
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {"name": "test_2"},
        },
    ]

    # when
    results = merge_documents(metadata_list)

    # then
    expected_results = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "author": [
            {"name": "test_1"},
            {"name": "test_2"},
        ],
    }
    assert results == expected_results



# Copyright (C) 2015-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict, Iterable, Iterator, List

from swh.core.db import BaseDb
from swh.core.db.db_utils import execute_values_generator, stored_procedure

from .interface import Sha1


class Db(BaseDb):
    """Proxy to the SWH Indexer DB, with wrappers around stored procedures"""

    content_mimetype_hash_keys = ["id", "indexer_configuration_id"]

    def _missing_from_list(
        self, table: str, data: Iterable[Dict], hash_keys: List[str], cur=None
    ):
        """Read from table the data with hash_keys that are missing.

        Args:
            table: Table name (e.g content_mimetype, fossology_license, etc...)
            data: Dict of data to read from
            hash_keys: List of keys to read in the data dict.

        Yields:
            The data which is missing from the db.

        """
        cur = self._cursor(cur)
        keys = ", ".join(hash_keys)
        equality = " AND ".join(("t.%s = c.%s" % (key, key)) for key in hash_keys)
        yield from execute_values_generator(
            cur,
            """
            select %s from (values %%s) as t(%s)
            where not exists (
                select 1 from %s c
                where %s
            )
            """
            % (keys, keys, table, equality),
            (tuple(m[k] for k in hash_keys) for m in data),
        )

    def content_mimetype_missing_from_list(
        self, mimetypes: Iterable[Dict], cur=None
    ) -> Iterator[Sha1]:
        """List missing mimetypes."""
        yield from self._missing_from_list(
            "content_mimetype", mimetypes, self.content_mimetype_hash_keys, cur=cur
        )

    content_mimetype_cols = [
        "id",
        "mimetype",
        "encoding",
        "tool_id",
        "tool_name",
        "tool_version",
        "tool_configuration",
    ]

    @stored_procedure("swh_mktemp_content_mimetype")
    def mktemp_content_mimetype(self, cur=None):
        pass

    def content_mimetype_add_from_temp(self, cur=None):
        cur = self._cursor(cur)
        cur.execute("select * from swh_content_mimetype_add()")
        return cur.fetchone()[0]

    def _convert_key(self, key, main_table="c"):
        """Convert keys according to specific use in the module.

        Args:
            key (str): Key expression to change according to the alias
              used in the query
            main_table (str): Alias to use for the main table. Default
              to c for content_{something}.

        Expected:
            Tables content_{something} being aliased as 'c' (something
            in {mimetype, ...}), table indexer_configuration being aliased as 'i'.

        """
        if key == "id":
            return "%s.id" % main_table
        elif key == "tool_id":
            return "i.id as tool_id"
        elif key == "license":
            return (
                """
                (
                    select name
                    from fossology_license
                    where id = %s.license_id
                )
                as licenses"""
                % main_table
            )
        return key

    def _get_from_list(self, table, ids, cols, cur=None, id_col="id"):
        """Fetches entries from the `table` such that their `id` field
        (or whatever is given to `id_col`) is in `ids`.
        Returns the columns `cols`.
        The `cur` parameter is used to connect to the database.
        """
        cur = self._cursor(cur)
        keys = map(self._convert_key, cols)
        query = """
            select {keys}
            from (values %s) as t(id)
            inner join {table} c
                on c.{id_col}=t.id
            inner join indexer_configuration i
                on c.indexer_configuration_id=i.id;
            """.format(
            keys=", ".join(keys), id_col=id_col, table=table
        )
        yield from execute_values_generator(cur, query, ((_id,) for _id in ids))

    content_indexer_names = {
        "mimetype": "content_mimetype",
        "fossology_license": "content_fossology_license",
    }

    def content_get_range(
        self,
        content_type,
        start,
        end,
        indexer_configuration_id,
        limit=1000,
        with_textual_data=False,
        cur=None,
    ):
        """Retrieve contents with content_type, within range [start, end]
        bound by limit and associated to the given indexer
        configuration id.

        When asking to work on textual content, that filters on the
        mimetype table with any mimetype that is not binary.

        """
        cur = self._cursor(cur)
        table = self.content_indexer_names[content_type]
        if with_textual_data:
            extra = """inner join content_mimetype cm
                         on (t.id=cm.id and cm.mimetype like 'text/%%' and
                             %(start)s <= cm.id and cm.id <= %(end)s)
                    """
        else:
            extra = ""
        query = f"""select t.id
                    from {table} t
                    {extra}
                    where t.indexer_configuration_id=%(tool_id)s
                          and %(start)s <= t.id and t.id <= %(end)s
                    order by t.indexer_configuration_id, t.id
                    limit %(limit)s"""
        cur.execute(
            query,
            {
                "start": start,
                "end": end,
                "tool_id": indexer_configuration_id,
                "limit": limit,
            },
        )
        yield from cur

    def content_mimetype_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            "content_mimetype", ids, self.content_mimetype_cols, cur=cur
        )

    content_fossology_license_cols = [
        "id",
        "tool_id",
        "tool_name",
        "tool_version",
        "tool_configuration",
        "license",
    ]

    @stored_procedure("swh_mktemp_content_fossology_license")
    def mktemp_content_fossology_license(self, cur=None):
        pass

    def content_fossology_license_add_from_temp(self, cur=None):
        """Add new licenses per content."""
        cur = self._cursor(cur)
        cur.execute("select * from swh_content_fossology_license_add()")
        return cur.fetchone()[0]

    def content_fossology_license_get_from_list(self, ids, cur=None):
        """Retrieve licenses per id."""
        cur = self._cursor(cur)
        keys = map(self._convert_key, self.content_fossology_license_cols)
        yield from execute_values_generator(
            cur,
            """
            select %s
            from (values %%s) as t(id)
            inner join content_fossology_license c on t.id=c.id
            inner join indexer_configuration i
                on i.id=c.indexer_configuration_id
            """
            % ", ".join(keys),
            ((_id,) for _id in ids),
        )

    content_metadata_hash_keys = ["id", "indexer_configuration_id"]

    def content_metadata_missing_from_list(self, metadata, cur=None):
        """List missing metadata."""
        yield from self._missing_from_list(
            "content_metadata", metadata, self.content_metadata_hash_keys, cur=cur
        )

    content_metadata_cols = [
        "id",
        "metadata",
        "tool_id",
        "tool_name",
        "tool_version",
        "tool_configuration",
    ]

    @stored_procedure("swh_mktemp_content_metadata")
    def mktemp_content_metadata(self, cur=None):
        pass

    def content_metadata_add_from_temp(self, cur=None):
        cur = self._cursor(cur)
        cur.execute("select * from swh_content_metadata_add()")
        return cur.fetchone()[0]

    def content_metadata_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            "content_metadata", ids, self.content_metadata_cols, cur=cur
        )

    directory_intrinsic_metadata_hash_keys = ["id", "indexer_configuration_id"]

    def directory_intrinsic_metadata_missing_from_list(self, metadata, cur=None):
        """List missing metadata."""
        yield from self._missing_from_list(
            "directory_intrinsic_metadata",
            metadata,
            self.directory_intrinsic_metadata_hash_keys,
            cur=cur,
        )

    directory_intrinsic_metadata_cols = [
        "id",
        "metadata",
        "mappings",
        "tool_id",
        "tool_name",
        "tool_version",
        "tool_configuration",
    ]

    @stored_procedure("swh_mktemp_directory_intrinsic_metadata")
    def mktemp_directory_intrinsic_metadata(self, cur=None):
        pass

    def directory_intrinsic_metadata_add_from_temp(self, cur=None):
        cur = self._cursor(cur)
        cur.execute("select * from swh_directory_intrinsic_metadata_add()")
        return cur.fetchone()[0]

    def directory_intrinsic_metadata_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            "directory_intrinsic_metadata",
            ids,
            self.directory_intrinsic_metadata_cols,
            cur=cur,
        )

    origin_intrinsic_metadata_cols = [
        "id",
        "metadata",
        "from_directory",
        "mappings",
        "tool_id",
        "tool_name",
        "tool_version",
        "tool_configuration",
    ]

    origin_intrinsic_metadata_regconfig = "pg_catalog.simple"
    """The dictionary used to normalize 'metadata' and queries.
    'pg_catalog.simple' provides no stopword, so it should be suitable
    for proper names and non-English content.
    When updating this value, make sure to add a new index on
    origin_intrinsic_metadata.metadata."""

    @stored_procedure("swh_mktemp_origin_intrinsic_metadata")
    def mktemp_origin_intrinsic_metadata(self, cur=None):
        pass

    def origin_intrinsic_metadata_add_from_temp(self, cur=None):
        cur = self._cursor(cur)
        cur.execute("select * from swh_origin_intrinsic_metadata_add()")
        return cur.fetchone()[0]

    def origin_intrinsic_metadata_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            "origin_intrinsic_metadata",
            ids,
            self.origin_intrinsic_metadata_cols,
            cur=cur,
            id_col="id",
        )

    def origin_intrinsic_metadata_search_fulltext(self, terms, *, limit, cur):
        regconfig = self.origin_intrinsic_metadata_regconfig
        tsquery_template = " && ".join(
            "plainto_tsquery('%s', %%s)" % regconfig for _ in terms
        )
        tsquery_args = [(term,) for term in terms]
        keys = (
            self._convert_key(col, "oim") for col in self.origin_intrinsic_metadata_cols
        )

        query = (
            "SELECT {keys} FROM origin_intrinsic_metadata AS oim "
            "INNER JOIN indexer_configuration AS i "
            "ON oim.indexer_configuration_id=i.id "
            "JOIN LATERAL (SELECT {tsquery_template}) AS s(tsq) ON true "
            "WHERE oim.metadata_tsvector @@ tsq "
            "ORDER BY ts_rank(oim.metadata_tsvector, tsq, 1) DESC "
            "LIMIT %s;"
        ).format(keys=", ".join(keys), tsquery_template=tsquery_template)
        cur.execute(query, tsquery_args + [limit])
        yield from cur

    def origin_intrinsic_metadata_search_by_producer(
        self, last, limit, ids_only, mappings, tool_ids, cur
    ):
        if ids_only:
            keys = "oim.id"
        else:
            keys = ", ".join(
                (
                    self._convert_key(col, "oim")
                    for col in self.origin_intrinsic_metadata_cols
                )
            )
        query_parts = [
            "SELECT %s" % keys,
            "FROM origin_intrinsic_metadata AS oim",
            "INNER JOIN indexer_configuration AS i",
            "ON oim.indexer_configuration_id=i.id",
        ]
        args = []

        where = []
        if last:
            where.append("oim.id > %s")
            args.append(last)
        if mappings is not None:
            where.append("oim.mappings && %s")
            args.append(list(mappings))
        if tool_ids is not None:
            where.append("oim.indexer_configuration_id = ANY(%s)")
            args.append(list(tool_ids))
        if where:
            query_parts.append("WHERE")
            query_parts.append(" AND ".join(where))

        if limit:
            query_parts.append("LIMIT %s")
            args.append(limit)

        cur.execute(" ".join(query_parts), args)
        yield from cur

    origin_extrinsic_metadata_cols = [
        "id",
        "metadata",
        "from_remd_id",
        "mappings",
        "tool_id",
        "tool_name",
        "tool_version",
        "tool_configuration",
    ]

    @stored_procedure("swh_mktemp_origin_extrinsic_metadata")
    def mktemp_origin_extrinsic_metadata(self, cur=None):
        pass

    def origin_extrinsic_metadata_add_from_temp(self, cur=None):
        cur = self._cursor(cur)
        cur.execute("select * from swh_origin_extrinsic_metadata_add()")
        return cur.fetchone()[0]

    def origin_extrinsic_metadata_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            "origin_extrinsic_metadata",
            ids,
            self.origin_extrinsic_metadata_cols,
            cur=cur,
            id_col="id",
        )

    indexer_configuration_cols = [
        "id",
        "tool_name",
        "tool_version",
        "tool_configuration",
    ]

    @stored_procedure("swh_mktemp_indexer_configuration")
    def mktemp_indexer_configuration(self, cur=None):
        pass

    def indexer_configuration_add_from_temp(self, cur=None):
        cur = self._cursor(cur)
        cur.execute(
            "SELECT %s from swh_indexer_configuration_add()"
            % (",".join(self.indexer_configuration_cols),)
        )
        yield from cur

    def indexer_configuration_get(
        self, tool_name, tool_version, tool_configuration, cur=None
    ):
        cur = self._cursor(cur)
        cur.execute(
            """select %s
                       from indexer_configuration
                       where tool_name=%%s and
                             tool_version=%%s and
                             tool_configuration=%%s"""
            % (",".join(self.indexer_configuration_cols)),
            (tool_name, tool_version, tool_configuration),
        )

        return cur.fetchone()

    def indexer_configuration_get_from_id(self, id_, cur=None):
        cur = self._cursor(cur)
        cur.execute(
            """select %s
                       from indexer_configuration
                       where id=%%s"""
            % (",".join(self.indexer_configuration_cols)),
            (id_,),
        )

        return cur.fetchone()
